"""
<Program Name>
  util.py

<Author>
  Konstantin Andrianov

<Started>
  March 24, 2012.  Derived from original util.py written by Geremy Condra.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides utility services.  This module supplies utility functions such as:
  get_file_details() that computes the length and hash of a file, import_json
  that tries to import a working json module, load_json_* functions, and a
  TempFile class that generates a file-like object for temporary storage, etc.
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import sys
import gzip
import shutil
import logging
import tempfile
import fnmatch

import securesystemslib.exceptions
import securesystemslib.settings
import securesystemslib.hash
import securesystemslib.formats

import six

# The algorithm used by the repository to generate the digests of the
# target filepaths, which are included in metadata files and may be prepended
# to the filenames of consistent snapshots.
HASH_FUNCTION = 'sha256'

# See 'log.py' to learn how logging is handled in TUF.
logger = logging.getLogger('securesystemslib_util')


class TempFile(object):
  """
  <Purpose>
    A high-level temporary file that cleans itself up or can be manually
    cleaned up. This isn't a complete file-like object. The file functions
    that are supported make additional common-case safe assumptions.  There
    are additional functions that aren't part of file-like objects.  TempFile
    is used in the download.py module to temporarily store downloaded data while
    all security checks (file hashes/length) are performed.
  """

  def _default_temporary_directory(self, prefix):
    """__init__ helper."""
    try:
      self.temporary_file = tempfile.NamedTemporaryFile(prefix=prefix)

    except OSError as err: # pragma: no cover
      logger.critical('Cannot create a system temporary directory: '+repr(err))
      raise securesystemslib.exceptions.Error(err)


  def __init__(self, prefix='tuf_temp_'):
    """
    <Purpose>
      Initializes TempFile.

    <Arguments>
      prefix:
        A string argument to be used with tempfile.NamedTemporaryFile function.

    <Exceptions>
      securesystemslib.exceptions.Error on failure to load temp dir.

    <Return>
      None.
    """

    self._compression = None

    # If compression is set then the original file is saved in 'self._orig_file'.
    self._orig_file = None
    temp_dir = securesystemslib.settings.temporary_directory
    if temp_dir is not None and securesystemslib.formats.PATH_SCHEMA.matches(temp_dir):
      try:
        self.temporary_file = tempfile.NamedTemporaryFile(prefix=prefix,
                                                          dir=temp_dir)
      except OSError as err:
        logger.error('Temp file in ' + temp_dir + ' failed: ' +repr(err))
        logger.error('Will attempt to use system default temp dir.')
        self._default_temporary_directory(prefix)

    else:
      self._default_temporary_directory(prefix)


  def get_compressed_length(self):
    """
    <Purpose>
      Get the compressed length of the file. This will be correct information
      even when the file is read as an uncompressed one.

    <Arguments>
      None.

    <Exceptions>
      OSError.

    <Return>
      Nonnegative integer representing compressed file size.
    """

    # Even if we read a compressed file with the gzip standard library module,
    # the original file will remain compressed.
    return os.stat(self.temporary_file.name).st_size


  def flush(self):
    """
    <Purpose>
      Flushes buffered output for the file.

    <Arguments>
      None.

    <Exceptions>
      None.

    <Return>
      None.
    """

    self.temporary_file.flush()


  def read(self, size=None):
    """
    <Purpose>
      Read specified number of bytes.  If size is not specified then the whole
      file is read and the file pointer is placed at the beginning of the file.

    <Arguments>
      size:
        Number of bytes to be read.

    <Exceptions>
      securesystemslib.exceptions.FormatError: if 'size' is invalid.

    <Return>
      String of data.
    """

    if size is None:
      self.temporary_file.seek(0)
      data = self.temporary_file.read()
      self.temporary_file.seek(0)

      return data

    else:
      if not (isinstance(size, int) and size > 0):
        raise securesystemslib.exceptions.FormatError

      return self.temporary_file.read(size)


  def write(self, data, auto_flush=True):
    """
    <Purpose>
      Writes a data string to the file.

    <Arguments>
      data:
        A string containing some data.

      auto_flush:
        Boolean argument, if set to 'True', all data will be flushed from
        internal buffer.

    <Exceptions>
      None.

    <Return>
      None.
    """

    self.temporary_file.write(data)
    if auto_flush:
      self.flush()


  def move(self, destination_path):
    """
    <Purpose>
      Copies 'self.temporary_file' to a non-temp file at 'destination_path' and
      closes 'self.temporary_file' so that it is removed.

    <Arguments>
      destination_path:
        Path to store the file in.

    <Exceptions>
      None.

    <Return>
      None.
    """

    self.flush()
    self.seek(0)
    destination_file = open(destination_path, 'wb')
    shutil.copyfileobj(self.temporary_file, destination_file)
    # Force the destination file to be written to disk from Python's internal
    # and the operation system's buffers.  os.fsync() should follow flush().
    destination_file.flush()
    os.fsync(destination_file.fileno())
    destination_file.close()

    # 'self.close()' closes temporary file which destroys itself.
    self.close_temp_file()


  def seek(self, *args):
    """
    <Purpose>
      Set file's current position.

    <Arguments>
      *args:
        (*-operator): unpacking argument list is used because seek method
        accepts two args: offset and whence.  If whence is not specified, its
        default is 0.  Indicate offset to set the file's current position.
        Refer to the python manual for more info.

    <Exceptions>
      None.

    <Return>
      None.
    """

    self.temporary_file.seek(*args)


  def decompress_temp_file_object(self, compression):
    """
    <Purpose>
      To decompress a compressed temp file object.  Decompression is performed
      on a temp file object that is compressed, this occurs after downloading
      a compressed file.  For instance if a compressed version of some meta
      file in the repository is downloaded, the temp file containing the
      compressed meta file will be decompressed using this function.
      Note that after calling this method, write() can no longer be called.

                            meta.json.gz
                               |...[download]
                        temporary_file (containing meta.json.gz)
                        /             \
               temporary_file          _orig_file
          containing meta.json          containing meta.json.gz
          (decompressed data)

    <Arguments>
      compression:
        A string indicating the type of compression that was used to compress
        a file.  Only gzip is allowed.

    <Exceptions>
      securesystemslib.exceptions.FormatError: If 'compression' is improperly formatted.

      securesystemslib.exceptions.Error: If an invalid compression is given.

      securesystemslib.exceptions.DecompressionError: If the compression failed for any reason.

    <Side Effects>
      'self._orig_file' is used to store the original data of 'temporary_file'.

    <Return>
      None.
    """

    # Does 'compression' have the correct format?
    # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
    securesystemslib.formats.NAME_SCHEMA.check_match(compression)

    if self._orig_file is not None:
      raise securesystemslib.exceptions.Error('Can only set compression on a'
          ' TempFile once.')

    if compression != 'gzip':
      raise securesystemslib.exceptions.Error('Only gzip compression is'
          ' supported.')

    self.seek(0)
    self._compression = compression
    self._orig_file = self.temporary_file

    try:
      gzip_file_object = gzip.GzipFile(fileobj=self.temporary_file, mode='rb')
      uncompressed_content = gzip_file_object.read()
      self.temporary_file = tempfile.NamedTemporaryFile()
      self.temporary_file.write(uncompressed_content)
      self.flush()

    except Exception as exception:
      raise securesystemslib.exceptions.DecompressionError(exception)


  def close_temp_file(self):
    """
    <Purpose>
      Closes the temporary file object. 'close_temp_file' mimics usual
      file.close(), however temporary file destroys itself when
      'close_temp_file' is called. Further if compression is set, second
      temporary file instance 'self._orig_file' is also closed so that no open
      temporary files are left open.

    <Arguments>
      None.

    <Exceptions>
      None.

    <Side Effects>
      Closes 'self._orig_file'.

    <Return>
      None.
    """

    self.temporary_file.close()
    # If compression has been set, we need to explicitly close the original
    # file object.
    if self._orig_file is not None:
      self._orig_file.close()


def get_file_details(filepath, hash_algorithms=['sha256']):
  """
  <Purpose>
    To get file's length and hash information.  The hash is computed using the
    sha256 algorithm.  This function is used in the signerlib.py and updater.py
    modules.

  <Arguments>
    filepath:
      Absolute file path of a file.

    hash_algorithms:

  <Exceptions>
    securesystemslib.exceptions.FormatError: If hash of the file does not match
    HASHDICT_SCHEMA.

    securesystemslib.exceptions.Error: If 'filepath' does not exist.

  <Returns>
    A tuple (length, hashes) describing 'filepath'.
  """

  # Making sure that the format of 'filepath' is a path string.
  # 'securesystemslib.exceptions.FormatError' is raised on incorrect format.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)
  securesystemslib.formats.HASHALGORITHMS_SCHEMA.check_match(hash_algorithms)

  # The returned file hashes of 'filepath'.
  file_hashes = {}

  # Does the path exists?
  if not os.path.exists(filepath):
    raise securesystemslib.exceptions.Error('Path ' + repr(filepath) + ' doest'
        ' not exist.')

  filepath = os.path.abspath(filepath)

  # Obtaining length of the file.
  file_length = os.path.getsize(filepath)

  # Obtaining hash of the file.
  for algorithm in hash_algorithms:
    digest_object = securesystemslib.hash.digest_filename(filepath, algorithm)
    file_hashes.update({algorithm: digest_object.hexdigest()})

  # Performing a format check to ensure 'file_hash' corresponds HASHDICT_SCHEMA.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.HASHDICT_SCHEMA.check_match(file_hashes)

  return file_length, file_hashes


def ensure_parent_dir(filename):
  """
  <Purpose>
    To ensure existence of the parent directory of 'filename'.  If the parent
    directory of 'name' does not exist, create it.

    Example: If 'filename' is '/a/b/c/d.txt', and only the directory '/a/b/'
    exists, then directory '/a/b/c/d/' will be created.

  <Arguments>
    filename:
      A path string.

  <Exceptions>
    securesystemslib.exceptions.FormatError: If 'filename' is improperly
    formatted.

  <Side Effects>
    A directory is created whenever the parent directory of 'filename' does not
    exist.

  <Return>
    None.
  """

  # Ensure 'filename' corresponds to 'PATH_SCHEMA'.
  # Raise 'securesystemslib.exceptions.FormatError' on a mismatch.
  securesystemslib.formats.PATH_SCHEMA.check_match(filename)

  # Split 'filename' into head and tail, check if head exists.
  directory = os.path.split(filename)[0]

  if directory and not os.path.exists(directory):
    # mode = 'rwx------'. 448 (decimal) is 700 in octal.
    os.makedirs(directory, 448)


def file_in_confined_directories(filepath, confined_directories):
  """
  <Purpose>
    Check if the directory containing 'filepath' is in the list/tuple of
    'confined_directories'.

  <Arguments>
    filepath:
      A string representing the path of a file.  The following example path
      strings are viewed as files and not directories: 'a/b/c', 'a/b/c.txt'.

    confined_directories:
      A list, or a tuple, of directory strings.

  <Exceptions>
   securesystemslib.exceptions.FormatError: On incorrect format of the input.

  <Return>
    Boolean.  True, if path is either the empty string
    or in 'confined_paths'; False, otherwise.
  """

  # Do the arguments have the correct format?
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.RELPATH_SCHEMA.check_match(filepath)
  securesystemslib.formats.RELPATHS_SCHEMA.check_match(confined_directories)

  for confined_directory in confined_directories:
    # The empty string (arbitrarily chosen) signifies the client is confined
    # to all directories and subdirectories.  No need to check 'filepath'.
    if confined_directory == '':
      return True

    # Normalized paths needed, to account for up-level references, etc.
    # TUF clients have the option of setting the list of directories in
    # 'confined_directories'.
    filepath = os.path.normpath(filepath)
    confined_directory = os.path.normpath(confined_directory)

    # A TUF client may restrict himself to specific directories on the
    # remote repository.  The list of paths in 'confined_path', not including
    # each path's subdirectories, are the only directories the client will
    # download targets from.
    if os.path.dirname(filepath) == confined_directory:
      return True

  return False


def find_delegated_role(roles, delegated_role):
  """
  <Purpose>
    Find the index, if any, of a role with a given name in a list of roles.

  <Arguments>
    roles:
      The list of roles, each of which must have a 'name' attribute.

    delegated_role:
      The name of the role to be found in the list of roles.

  <Exceptions>
    securesystemslib.exceptions.RepositoryError, if the list of roles has
    invalid data.

  <Side Effects>
    No known side effects.

  <Returns>
    The unique index, an interger, in the list of roles.  if 'delegated_role'
    does not exist, 'None' is returned.
  """

  # Do the arguments have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.  Raise
  # 'securesystemslib.exceptions.FormatError' if any are improperly formatted.
  securesystemslib.formats.ROLELIST_SCHEMA.check_match(roles)
  securesystemslib.formats.ROLENAME_SCHEMA.check_match(delegated_role)

  # The index of a role, if any, with the same name.
  role_index = None

  for index in six.moves.xrange(len(roles)):
    role = roles[index]
    name = role.get('name')

    # This role has no name.
    if name is None:
      no_name_message = 'Role with no name.'
      raise securesystemslib.exceptions.RepositoryError(no_name_message)

    # Does this role have the same name?
    else:
      # This role has the same name, and...
      if name == delegated_role:
        # ...it is the only known role with the same name.
        if role_index is None:
          role_index = index

        # ...there are at least two roles with the same name.
        else:
          duplicate_role_message = 'Duplicate role (' + str(delegated_role) + ').'
          raise securesystemslib.exceptions.RepositoryError(
              'Duplicate role (' + str(delegated_role) + ').')

      # This role has a different name.
      else:
        logger.debug('Skipping delegated role: ' + repr(delegated_role))

  return role_index


def ensure_all_targets_allowed(rolename, list_of_targets, parent_delegations):
  """
  <Purpose>
    Ensure that the list of targets specified by 'rolename' are allowed; this
    is determined by inspecting the 'delegations' field of the parent role of
    'rolename'.  If a target specified by 'rolename' is not found in the
    delegations field of 'metadata_object_of_parent', raise an exception.  The
    top-level role 'targets' is allowed to list any target file, so this
    function does not raise an exception if 'rolename' is 'targets'.

    Targets allowed are either exlicitly listed under the 'paths' field, or
    implicitly exist under a subdirectory of a parent directory listed under
    'paths'.  A parent role may delegate trust to all files under a particular
    directory, including files in subdirectories, by simply listing the
    directory (e.g., '/packages/source/Django/', the equivalent of
    '/packages/source/Django/*').  Targets listed in hashed bins are also
    validated (i.e., its calculated path hash prefix must be delegated by the
    parent role).

    TODO: Should the TUF spec restrict the repository to one particular
    algorithm when calcutating path hash prefixes (currently restricted to
    SHA256)?  Should we allow the repository to specify in the role dictionary
    the algorithm used for these generated hashed paths?

  <Arguments>
    rolename:
      The name of the role whose targets must be verified. This is a
      role name and should not end in '.json'.  Examples: 'root', 'targets',
      'targets/linux/x86'.

    list_of_targets:
      The targets of 'rolename', as listed in targets field of the 'rolename'
      metadata.  'list_of_targets' are target paths relative to the targets
      directory of the repository.  The delegations of the parent role are
      checked to verify that the targets of 'list_of_targets' are valid.

    parent_delegations:
      The parent delegations of 'rolename'.  The metadata object stores
      the allowed paths and path hash prefixes of child delegations in its
      'delegations' attribute.

  <Exceptions>
    securesystemslib.exceptions.FormatError:
      If any of the arguments are improperly formatted.

    securesystemslib.exceptions.ForbiddenTargetError:
      If the targets of 'metadata_role' are not allowed according to
      the parent's metadata file.  The 'paths' and 'path_hash_prefixes'
      attributes are verified.

    securesystemslib.exceptions.RepositoryError:
      If the parent of 'rolename' has not made a delegation to 'rolename'.

  <Side Effects>
    None.

  <Returns>
    None.
  """

  # Do the arguments have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.  Raise
  # 'securesystemslib.exceptions.FormatError' if any are improperly formatted.
  securesystemslib.formats.ROLENAME_SCHEMA.check_match(rolename)
  securesystemslib.formats.RELPATHS_SCHEMA.check_match(list_of_targets)
  securesystemslib.formats.DELEGATIONS_SCHEMA.check_match(parent_delegations)

  # Return if 'rolename' is 'targets'.  'targets' is not a delegated role.  Any
  # target file listed in 'targets' is allowed.
  if rolename == 'targets':
    return

  # The allowed targets of delegated roles are stored in the parent's metadata
  # file.  Iterate 'list_of_targets' and confirm they are trusted, or their
  # root parent directory exists in the role delegated paths, or path hash
  # prefixes, of the parent role.  First, locate 'rolename' in the 'roles'
  # attribute of 'parent_delegations'.
  roles = parent_delegations['roles']
  role_index = find_delegated_role(roles, rolename)

  # Ensure the delegated role exists prior to extracting trusted paths from
  # the parent's 'paths', or trusted path hash prefixes from the parent's
  # 'path_hash_prefixes'.
  if role_index is not None:
    role = roles[role_index]
    allowed_child_paths = role.get('paths')
    allowed_child_path_hash_prefixes = role.get('path_hash_prefixes')
    actual_child_targets = list_of_targets

    if allowed_child_path_hash_prefixes is not None:
      consistent = paths_are_consistent_with_hash_prefixes

      # 'actual_child_tarets' (i.e., 'list_of_targets') should have lenth
      # greater than zero due to the format check above.
      if not consistent(actual_child_targets,
                        allowed_child_path_hash_prefixes):
        message =  repr(rolename) + ' specifies a target that does not' + \
          ' have a path hash prefix listed in its parent role.'
        raise securesystemslib.exceptions.ForbiddenTargetError(message)

    elif allowed_child_paths is not None:
      # Check that each delegated target is either explicitly listed or a
      # parent directory is found under role['paths'], otherwise raise an
      # exception.  If the parent role explicitly lists target file paths in
      # 'paths', this loop will run in O(n^2), the worst-case.  The repository
      # maintainer will likely delegate entire directories, and opt for
      # explicit file paths if the targets in a directory are delegated to
      # different roles/developers.
      for child_target in actual_child_targets:
        for allowed_child_path in allowed_child_paths:
          if fnmatch.fnmatch(child_target, allowed_child_path):
            break

        else:
          raise securesystemslib.exceptions.ForbiddenTargetError(
              'Role ' + repr(rolename) + ' specifies'
              ' target' + repr(child_target) + ',' + ' which is not an allowed'
              ' path according to the delegations set by its parent role.')

    else:
      # 'role' should have been validated when it was downloaded.
      # The 'paths' or 'path_hash_prefixes' attributes should not be missing,
      # so raise an error in case this clause is reached.
      raise securesystemslib.exceptions.FormatError(repr(role) + ' did not'
          ' contain one of the required fields ("paths" or'
          ' "path_hash_prefixes").')

  # Raise an exception if the parent has not delegated to the specified
  # 'rolename' child role.
  else:
    raise securesystemslib.exceptions.RepositoryError('The parent role has'
        ' not delegated to ' + repr(rolename) + '.')


def paths_are_consistent_with_hash_prefixes(paths, path_hash_prefixes):
  """
  <Purpose>
    Determine whether a list of paths are consistent with their alleged path
    hash prefixes. By default, the SHA256 hash function is used.

  <Arguments>
    paths:
      A list of paths for which their hashes will be checked.

    path_hash_prefixes:
      The list of path hash prefixes with which to check the list of paths.

  <Exceptions>
    securesystemslib.exceptions.FormatError:
      If the arguments are improperly formatted.

  <Side Effects>
    No known side effects.

  <Returns>
    A Boolean indicating whether or not the paths are consistent with the
    hash prefix.
  """

  # Do the arguments have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.  Raise
  # 'securesystemslib.exceptions.FormatError' if any are improperly formatted.
  securesystemslib.formats.RELPATHS_SCHEMA.check_match(paths)
  securesystemslib.formats.PATH_HASH_PREFIXES_SCHEMA.check_match(path_hash_prefixes)

  # Assume that 'paths' and 'path_hash_prefixes' are inconsistent until
  # proven otherwise.
  consistent = False

  # The format checks above ensure the 'paths' and 'path_hash_prefix' lists
  # have lengths greater than zero.
  for path in paths:
    path_hash = get_target_hash(path)

    # Assume that every path is inconsistent until proven otherwise.
    consistent = False

    for path_hash_prefix in path_hash_prefixes:
      if path_hash.startswith(path_hash_prefix):
        consistent = True
        break

    # This path has no matching path_hash_prefix. Stop looking further.
    if not consistent:
      break

  return consistent


def get_target_hash(target_filepath):
  """
  <Purpose>
    Compute the hash of 'target_filepath'. This is useful in conjunction with
    the "path_hash_prefixes" attribute in a delegated targets role, which tells
    us which paths it is implicitly responsible for.

    The repository may optionally organize targets into hashed bins to ease
    target delegations and role metadata management.  The use of consistent
    hashing allows for a uniform distribution of targets into bins.

  <Arguments>
    target_filepath:
      The path to the target file on the repository. This will be relative to
      the 'targets' (or equivalent) directory on a given mirror.

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    The hash of 'target_filepath'.
  """

  # Does 'target_filepath' have the correct format?
  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.RELPATH_SCHEMA.check_match(target_filepath)

  # Calculate the hash of the filepath to determine which bin to find the
  # target.  The client currently assumes the repository uses
  # 'HASH_FUNCTION' to generate hashes and 'utf-8'.
  digest_object = securesystemslib.hash.digest(HASH_FUNCTION)
  encoded_target_filepath = target_filepath.encode('utf-8')
  digest_object.update(encoded_target_filepath)
  target_filepath_hash = digest_object.hexdigest()

  return target_filepath_hash


_json_module = None

def import_json():
  """
  <Purpose>
    Tries to import json module. We used to fall back to the simplejson module,
    but we have dropped support for that module. We are keeping this interface
    intact for backwards compatibility.

  <Arguments>
    None.

  <Exceptions>
    ImportError: on failure to import the json module.

  <Side Effects>
    None.

  <Return>
    json module
  """

  global _json_module

  if _json_module is not None:
    return _json_module

  else:
    try:
      module = __import__('json')

    # The 'json' module is available in Python > 2.6, and thus this exception
    # should not occur in all supported Python installations (> 2.6) of TUF.
    except ImportError: #pragma: no cover
      raise ImportError('Could not import the json module')

    else:
      _json_module = module
      return module

json = import_json()


def load_json_string(data):
  """
  <Purpose>
    Deserialize 'data' (JSON string) to a Python object.

  <Arguments>
    data:
      A JSON string.

  <Exceptions>
    securesystemslib.exceptions.Error, if 'data' cannot be deserialized to a
    Python object.

  <Side Effects>
    None.

  <Returns>
    Deserialized object.  For example, a dictionary.
  """

  deserialized_object = None

  try:
    deserialized_object = json.loads(data)

  except TypeError:
    message = 'Invalid JSON string: ' + repr(data)
    raise securesystemslib.exceptions.Error(message)

  except ValueError:
    message = 'Cannot deserialize to a Python object: ' + repr(data)
    raise securesystemslib.exceptions.Error(message)

  else:
    return deserialized_object


def load_json_file(filepath):
  """
  <Purpose>
    Deserialize a JSON object from a file containing the object.

  <Arguments>
    filepath:
      Absolute path of JSON file.

  <Exceptions>
    securesystemslib.exceptions.FormatError: If 'filepath' is improperly
    formatted.

    securesystemslib.exceptions.Error: If 'filepath' cannot be deserialized to
    a Python object.

    IOError in case of runtime IO exceptions.

  <Side Effects>
    None.

  <Return>
    Deserialized object.  For example, a dictionary.
  """

  # Making sure that the format of 'filepath' is a path string.
  # securesystemslib.exceptions.FormatError is raised on incorrect format.
  securesystemslib.formats.PATH_SCHEMA.check_match(filepath)

  deserialized_object = None

  # The file is mostly likely gzipped.
  if filepath.endswith('.gz'):
    logger.debug('gzip.open(' + str(filepath) + ')')
    fileobject = six.StringIO(gzip.open(filepath).read().decode('utf-8'))

  else:
    logger.debug('open(' + str(filepath) + ')')
    fileobject = open(filepath)

  try:
    deserialized_object = json.load(fileobject)

  except (ValueError, TypeError) as e:
    raise securesystemslib.exceptions.Error('Cannot deserialize to a'
      ' Python object: ' + repr(filepath))

  else:
    fileobject.close()
    return deserialized_object

  finally:
    fileobject.close()


def digests_are_equal(digest1, digest2):
  """
  <Purpose>
    While protecting against timing attacks, compare the hexadecimal arguments
    and determine if they are equal.

  <Arguments>
    digest1:
      The first hexadecimal string value to compare.

    digest2:
      The second hexadecimal string value to compare.

  <Exceptions>
    securesystemslib.exceptions.FormatError: If the arguments are improperly
    formatted.

  <Side Effects>
    None.

  <Return>
    Return True if 'digest1' is equal to 'digest2', False otherwise.
  """

  # Ensure the arguments have the appropriate number of objects and object
  # types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.HEX_SCHEMA.check_match(digest1)
  securesystemslib.formats.HEX_SCHEMA.check_match(digest2)

  if len(digest1) != len(digest2):
    return False

  are_equal = True

  for element in range(len(digest1)):
    if digest1[element] != digest2[element]:
      are_equal = False

  return are_equal



#!/usr/bin/env python

"""
<Program Name>
  test_util.py

<Author>
  Konstantin Andrianov.

<Started>
  February 1, 2013.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Unit test for 'util.py'
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import sys
import gzip
import shutil
import logging
import tempfile
import unittest
import timeit

import securesystemslib.settings
import securesystemslib.hash
import securesystemslib.util
import securesystemslib.unittest_toolbox as unittest_toolbox

import six

logger = logging.getLogger('securesystemslib_test_util')


class TestUtil(unittest_toolbox.Modified_TestCase):

  def setUp(self):
    unittest_toolbox.Modified_TestCase.setUp(self)
    self.temp_fileobj = securesystemslib.util.TempFile()

		

  def tearDown(self):
    unittest_toolbox.Modified_TestCase.tearDown(self)
    self.temp_fileobj.close_temp_file()



  def test_A1_tempfile_close_temp_file(self):
    # Was the temporary file closed?
    self.temp_fileobj.close_temp_file()
    self.assertTrue(self.temp_fileobj.temporary_file.closed)



  def _extract_tempfile_directory(self, config_temp_dir=None):
    """
      Takes a directory (essentially specified in the settings.py as
      'temporary_directory') and substitutes tempfile.TemporaryFile() with
      tempfile.mkstemp() in order to extract actual directory of the stored
      tempfile.  Returns the config's temporary directory (or default temp
      directory) and actual directory.
    """

    # Patching 'settings.temporary_directory'.
    securesystemslib.settings.temporary_directory = config_temp_dir

    if config_temp_dir is None:
      # 'config_temp_dir' needs to be set to default.
      config_temp_dir = tempfile.gettempdir()

    # Patching 'tempfile.TemporaryFile()' (by substituting
    # temfile.TemporaryFile() with tempfile.mkstemp()) in order to get the
    # directory of the stored tempfile object.
    saved_tempfile_TemporaryFile = securesystemslib.util.tempfile.NamedTemporaryFile
    securesystemslib.util.tempfile.NamedTemporaryFile = tempfile.mkstemp
    _temp_fileobj = securesystemslib.util.TempFile()
    securesystemslib.util.tempfile.NamedTemporaryFile = saved_tempfile_TemporaryFile
    junk, _tempfilepath = _temp_fileobj.temporary_file
    _tempfile_dir = os.path.dirname(_tempfilepath)

    # In the case when 'config_temp_dir' is None or some other discrepancy,
    # '_temp_fileobj' needs to be closed manually since tempfile.mkstemp()
    # was used.
    if os.path.exists(_tempfilepath):
      os.remove(_tempfilepath)

    return config_temp_dir, _tempfile_dir



  def test_A2_tempfile_init(self):
    # Goal: Verify that temporary files are stored in the appropriate temp
    # directory.  The location of the temporary files is set in 'settings.py'.

    # Test: Expected input verification.
    # Assumed 'settings.temporary_directory' is 'None' initially.
    temp_file = securesystemslib.util.TempFile()
    temp_file_directory = os.path.dirname(temp_file.temporary_file.name)
    self.assertEqual(tempfile.gettempdir(), temp_file_directory)

    saved_temporary_directory = securesystemslib.settings.temporary_directory
    temp_directory = self.make_temp_directory()
    securesystemslib.settings.temporary_directory = temp_directory
    temp_file = securesystemslib.util.TempFile()
    temp_file_directory = os.path.dirname(temp_file.temporary_file.name)
    self.assertEqual(temp_directory, temp_file_directory)

    securesystemslib.settings.temporary_directory = saved_temporary_directory

    # Test: Unexpected input handling.
    config_temp_dirs = [self.random_string(), 123, ['a'], {'a':1}]
    for config_temp_dir in config_temp_dirs:
      config_temp_dir, actual_dir = \
      self._extract_tempfile_directory(config_temp_dir)
      self.assertEqual(tempfile.gettempdir(), actual_dir)



  def test_A3_tempfile_read(self):
    filepath = self.make_temp_data_file(data = '1234567890')
    fileobj = open(filepath, 'rb')

    # Patching 'temp_fileobj.temporary_file'.
    self.temp_fileobj.temporary_file = fileobj

    # Test: Expected input.
    self.assertEqual(self.temp_fileobj.read().decode('utf-8'), '1234567890')
    self.assertEqual(self.temp_fileobj.read(4).decode('utf-8'), '1234')

    # Test: Unexpected input.
    for bogus_arg in ['abcd', ['abcd'], {'a':'a'}, -100]:
      self.assertRaises(securesystemslib.exceptions.FormatError,
          self.temp_fileobj.read, bogus_arg)



  def test_A4_tempfile_write(self):
    data = self.random_string()
    self.temp_fileobj.write(data.encode('utf-8'))
    self.assertEqual(data, self.temp_fileobj.read().decode('utf-8'))

    self.temp_fileobj.write(data.encode('utf-8'), auto_flush=False)
    self.assertEqual(data, self.temp_fileobj.read().decode('utf-8'))



  def test_A5_tempfile_move(self):
    # Destination directory to save the temporary file in.
    dest_temp_dir = self.make_temp_directory()
    dest_path = os.path.join(dest_temp_dir, self.random_string())
    self.temp_fileobj.write(self.random_string().encode('utf-8'))
    self.temp_fileobj.move(dest_path)
    self.assertTrue(dest_path)



  def _compress_existing_file(self, filepath):
    """
    [Helper]Compresses file 'filepath' and returns file path of
    the compresses file.
    """

    # NOTE: DO NOT forget to remove the newly created compressed file!
    if os.path.exists(filepath):
      compressed_filepath = filepath+'.gz'
      f_in = open(filepath, 'rb')
      f_out = gzip.open(compressed_filepath, 'wb')
      f_out.writelines(f_in)
      f_out.close()
      f_in.close()

      return compressed_filepath

    else:
      logger.error('Compression of ' + repr(filepath)  +' failed.'
          '  Path does not exist.')
      sys.exit(1)



  def _decompress_file(self, compressed_filepath):
    """[Helper]"""
    if os.path.exists(compressed_filepath):
      f = gzip.open(compressed_filepath, 'rb')
      file_content = f.read()
      f.close()
      return file_content

    else:
      logger.error('Decompression of ' + repr(compressed_filepath) + ' failed.'
          '  Path does not exist.')
      sys.exit(1)



  def test_A6_tempfile_decompress_temp_file_object(self):
    # Setup: generate a temp file (self.make_temp_data_file()),
    # compress it.  Write it to self.temp_fileobj().
    filepath = self.make_temp_data_file()
    fileobj = open(filepath, 'rb')
    compressed_filepath = self._compress_existing_file(filepath)
    compressed_fileobj = open(compressed_filepath, 'rb')
    self.temp_fileobj.write(compressed_fileobj.read())
    os.remove(compressed_filepath)

    # Try decompression using incorrect compression type i.e. compressions
    # other than 'gzip'.  In short feeding incorrect input.
    bogus_args = ['zip', 1234, self.random_string()]
    for arg in bogus_args:
      self.assertRaises(securesystemslib.exceptions.Error,
          self.temp_fileobj.decompress_temp_file_object, arg)

    # Test for a valid util.decompress_temp_file_object() call.
    self.temp_fileobj.decompress_temp_file_object('gzip')
    self.assertEqual(self.temp_fileobj.read(), fileobj.read())

    # Checking the content of the TempFile's '_orig_file' instance.
    check_compressed_original = self.make_temp_file()
    with open(check_compressed_original, 'wb') as file_object:
      self.temp_fileobj._orig_file.seek(0)
      original_content = self.temp_fileobj._orig_file.read()
      file_object.write(original_content)

    data_in_orig_file = self._decompress_file(check_compressed_original)
    fileobj.seek(0)
    self.assertEqual(data_in_orig_file, fileobj.read())

    # Try decompressing once more.
    self.assertRaises(securesystemslib.exceptions.Error,
        self.temp_fileobj.decompress_temp_file_object, 'gzip')

    # Test decompression of invalid gzip file.
    temp_file = securesystemslib.util.TempFile()
    temp_file.write(b'bad zip')
    contents = temp_file.read()
    self.assertRaises(securesystemslib.exceptions.DecompressionError,
        temp_file.decompress_temp_file_object, 'gzip')



  def test_B1_get_file_details(self):
    # Goal: Verify proper output given certain expected/unexpected input.

    # Making a temporary file.
    filepath = self.make_temp_data_file()

    # Computing the hash and length of the tempfile.
    digest_object = securesystemslib.hash.digest_filename(filepath, algorithm='sha256')
    file_hash = {'sha256' : digest_object.hexdigest()}
    file_length = os.path.getsize(filepath)

    # Test: Expected input.
    self.assertEqual(securesystemslib.util.get_file_details(filepath),
        (file_length, file_hash))

    # Test: Incorrect input.
    bogus_inputs = [self.random_string(), 1234, [self.random_string()],
        {'a': 'b'}, None]

    for bogus_input in bogus_inputs:
      if isinstance(bogus_input, six.string_types):
        self.assertRaises(securesystemslib.exceptions.Error,
            securesystemslib.util.get_file_details, bogus_input)
      else:
        self.assertRaises(securesystemslib.exceptions.FormatError,
            securesystemslib.util.get_file_details, bogus_input)



  def  test_B2_ensure_parent_dir(self):
    existing_parent_dir = self.make_temp_directory()
    non_existing_parent_dir = os.path.join(existing_parent_dir, 'a', 'b')

    for parent_dir in [existing_parent_dir, non_existing_parent_dir, 12, [3]]:
      if isinstance(parent_dir, six.string_types):
        securesystemslib.util.ensure_parent_dir(os.path.join(parent_dir, 'a.txt'))
        self.assertTrue(os.path.isdir(parent_dir))

      else:
        self.assertRaises(securesystemslib.exceptions.FormatError,
            securesystemslib.util.ensure_parent_dir, parent_dir)



  def  test_B3_file_in_confined_directories(self):
    # Goal: Provide invalid input for 'filepath' and 'confined_directories'.
    # Include inputs like: '[1, 2, "a"]' and such...
    # Reference to 'file_in_confined_directories()' to improve readability.
    in_confined_directory = securesystemslib.util.file_in_confined_directories
    list_of_confined_directories = ['a', 12, {'a':'a'}, [1]]
    list_of_filepaths = [12, ['a'], {'a':'a'}, 'a']
    for bogus_confined_directory in list_of_confined_directories:
      for filepath in list_of_filepaths:
        self.assertRaises(securesystemslib.exceptions.FormatError,
            in_confined_directory, filepath, bogus_confined_directory)

    # Test: Inputs that evaluate to False.
    confined_directories = ['a/b/', 'a/b/c/d/']
    self.assertFalse(in_confined_directory('a/b/c/1.txt', confined_directories))

    confined_directories = ['a/b/c/d/e/']
    self.assertFalse(in_confined_directory('a', confined_directories))
    self.assertFalse(in_confined_directory('a/b', confined_directories))
    self.assertFalse(in_confined_directory('a/b/c', confined_directories))
    self.assertFalse(in_confined_directory('a/b/c/d', confined_directories))
    # Below, 'e' is a file in the 'a/b/c/d/' directory.
    self.assertFalse(in_confined_directory('a/b/c/d/e', confined_directories))

    # Test: Inputs that evaluate to True.
    self.assertTrue(in_confined_directory('a/b/c.txt', ['']))
    self.assertTrue(in_confined_directory('a/b/c.txt', ['a/b/']))
    self.assertTrue(in_confined_directory('a/b/c.txt', ['x', '']))
    self.assertTrue(in_confined_directory('a/b/c/..', ['a/']))


  def test_B4_import_json(self):
    self.assertTrue('json' in sys.modules)
    json_module = securesystemslib.util.import_json()
    self.assertTrue(json_module is not None)

    # Test import_json() when 'util._json_moduel' is non-None.
    securesystemslib.util._json_module = 'junk_module'
    self.assertEqual(securesystemslib.util.import_json(), 'junk_module')



  def  test_B5_load_json_string(self):
    # Test normal case.
    data = ['a', {'b': ['c', None, 30.3, 29]}]
    json_string = securesystemslib.util.json.dumps(data)
    self.assertEqual(data, securesystemslib.util.load_json_string(json_string))

    # Test invalid arguments.
    self.assertRaises(securesystemslib.exceptions.Error,
        securesystemslib.util.load_json_string, 8)
    invalid_json_string = json_string + '.'
    self.assertRaises(securesystemslib.exceptions.Error,
        securesystemslib.util.load_json_string, invalid_json_string)



  def  test_B6_load_json_file(self):
    data = ['a', {'b': ['c', None, 30.3, 29]}]
    filepath = self.make_temp_file()
    fileobj = open(filepath, 'wt')
    securesystemslib.util.json.dump(data, fileobj)
    fileobj.close()
    self.assertEqual(data, securesystemslib.util.load_json_file(filepath))

    # Test a gzipped file.
    compressed_filepath = self._compress_existing_file(filepath)
    self.assertEqual(data,
        securesystemslib.util.load_json_file(compressed_filepath))

    # Improperly formatted arguments.
    for bogus_arg in [1, [b'a'], {'a':b'b'}]:
      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.util.load_json_file, bogus_arg)

    # Non-existent path.
    self.assertRaises(IOError,
        securesystemslib.util.load_json_file, 'non-existent.json')

    # Invalid JSON content.
    filepath_bad_data = self.make_temp_file()
    fileobj = open(filepath_bad_data, 'wt')
    fileobj.write('junk data')
    fileobj.close()

    self.assertRaises(securesystemslib.exceptions.Error,
      securesystemslib.util.load_json_file, filepath_bad_data)



  def test_C1_get_target_hash(self):
    # Test normal case.
    expected_target_hashes = {
      '/file1.txt': 'e3a3d89eb3b70ce3fbce6017d7b8c12d4abd5635427a0e8a238f53157df85b3d',
      '/README.txt': '8faee106f1bb69f34aaf1df1e3c2e87d763c4d878cb96b91db13495e32ceb0b0',
      '/warehouse/file2.txt': 'd543a573a2cec67026eff06e75702303559e64e705eba06f65799baaf0424417'
    }
    for filepath, target_hash in six.iteritems(expected_target_hashes):
      self.assertTrue(securesystemslib.formats.RELPATH_SCHEMA.matches(filepath))
      self.assertTrue(securesystemslib.formats.HASH_SCHEMA.matches(target_hash))
      self.assertEqual(securesystemslib.util.get_target_hash(filepath), target_hash)

    # Test for improperly formatted argument.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.get_target_hash, 8)



  def test_C2_find_delegated_role(self):
    # Test normal case.  Create an expected role list, which is one of the
    # required arguments to 'find_delegated_role()'.
    role_list = [
      {
       "keyids": [
        "a394c28384648328b16731f81440d72243c77bb44c07c040be99347f0df7d7bf"
       ],
       "name": "targets/warehouse",
       "paths": [
        "/file1.txt", "/README.txt", '/warehouse/'
       ],
       "threshold": 3
      },
      {
       "keyids": [
        "a394c28384648328b16731f81440d72243c77bb44c07c040be99347f0df7d7bf"
       ],
       "name": "targets/tuf",
       "paths": [
        "/updater.py", "formats.py", '/tuf/'
       ],
       "threshold": 4
      }
    ]

    self.assertTrue(securesystemslib.formats.ROLELIST_SCHEMA.matches(role_list))
    self.assertEqual(securesystemslib.util.find_delegated_role(role_list,
        'targets/tuf'), 1)
    self.assertEqual(securesystemslib.util.find_delegated_role(role_list,
        'targets/warehouse'), 0)

    # Test for non-existent role.  'find_delegated_role()' returns 'None'
    # if the role is not found.
    self.assertEqual(securesystemslib.util.find_delegated_role(role_list,
        'targets/non-existent'), None)

    # Test improperly formatted arguments.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.find_delegated_role, 8, role_list)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.find_delegated_role, 8, 'targets/tuf')

    # Test duplicate roles.
    role_list.append(role_list[1])
    self.assertRaises(securesystemslib.exceptions.RepositoryError,
        securesystemslib.util.find_delegated_role, role_list, 'targets/tuf')

    # Test missing 'name' attribute (optional, but required by
    # 'find_delegated_role()').
    # Delete the duplicate role, and the remaining role's 'name' attribute.
    del role_list[2]
    del role_list[0]['name']
    self.assertRaises(securesystemslib.exceptions.RepositoryError,
        securesystemslib.util.find_delegated_role, role_list,
        'targets/warehouse')



  def test_C3_paths_are_consistent_with_hash_prefixes(self):
    # Test normal case.
    path_hash_prefixes = ['e3a3', '8fae', 'd543']
    list_of_targets = ['/file1.txt', '/README.txt', '/warehouse/file2.txt']

    # Ensure the paths of 'list_of_targets' each have the expected path hash
    # prefix listed in 'path_hash_prefixes'.
    for filepath in list_of_targets:
      self.assertTrue(securesystemslib.util.get_target_hash(
          filepath)[0:4] in path_hash_prefixes)

    self.assertTrue(securesystemslib.util.paths_are_consistent_with_hash_prefixes(
        list_of_targets, path_hash_prefixes))

    extra_invalid_prefix = ['e3a3', '8fae', 'd543', '0000']
    self.assertTrue(securesystemslib.util.paths_are_consistent_with_hash_prefixes(
        list_of_targets, extra_invalid_prefix))

    # Test improperly formatted arguments.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.paths_are_consistent_with_hash_prefixes, 8,
        path_hash_prefixes)

    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.paths_are_consistent_with_hash_prefixes,
        list_of_targets, 8)

    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.paths_are_consistent_with_hash_prefixes,
        list_of_targets, ['zza1'])

    # Test invalid list of targets.
    bad_target_path = '/file5.txt'
    self.assertTrue(securesystemslib.util.get_target_hash(
        bad_target_path)[0:4] not in path_hash_prefixes)
    self.assertFalse(securesystemslib.util.paths_are_consistent_with_hash_prefixes(
        [bad_target_path], path_hash_prefixes))

    # Add invalid target path to 'list_of_targets'.
    list_of_targets.append(bad_target_path)
    self.assertFalse(securesystemslib.util.paths_are_consistent_with_hash_prefixes(
        list_of_targets, path_hash_prefixes))



  def test_C4_ensure_all_targets_allowed(self):
    # Test normal case.
    rolename = 'targets/warehouse'
    self.assertTrue(securesystemslib.formats.ROLENAME_SCHEMA.matches(rolename))
    list_of_targets = ['/file1.txt', '/README.txt', '/warehouse/file2.txt']
    self.assertTrue(securesystemslib.formats.RELPATHS_SCHEMA.matches(list_of_targets))
    parent_delegations = {"keys": {
      "a394c28384648328b16731f81440d72243c77bb44c07c040be99347f0df7d7bf": {
       "keytype": "ed25519",
       "scheme": "ed25519",
       "keyval": {
        "public": "3eb81026ded5af2c61fb3d4b272ac53cd1049a810ee88f4df1fc35cdaf918157"
       }
      }
     },
     "roles": [
      {
       "keyids": [
        "a394c28384648328b16731f81440d72243c77bb44c07c040be99347f0df7d7bf"
       ],
       "name": "targets/warehouse",
       "paths": [
        "/file*.txt", "/README.txt", '/warehouse/*'
       ],
       "threshold": 1
      }
     ]
    }
    self.assertTrue(securesystemslib.formats.DELEGATIONS_SCHEMA.matches(parent_delegations))

    securesystemslib.util.ensure_all_targets_allowed(rolename, list_of_targets,
        parent_delegations)

    # The target files of 'targets' are always allowed.  'list_of_targets' and
    # 'parent_delegations' are not checked in this case.
    securesystemslib.util.ensure_all_targets_allowed('targets', list_of_targets,
        parent_delegations)

    # Test improperly formatted arguments.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.ensure_all_targets_allowed, 8, list_of_targets,
        parent_delegations)

    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.ensure_all_targets_allowed, rolename, 8,
        parent_delegations)

    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.ensure_all_targets_allowed, rolename,
        list_of_targets, 8)

    # Test for invalid 'rolename', which has not been delegated by its parent,
    # 'targets'.
    self.assertRaises(securesystemslib.exceptions.RepositoryError,
        securesystemslib.util.ensure_all_targets_allowed,
        'targets/non-delegated_rolename', list_of_targets, parent_delegations)

    # Test for target file that is not allowed by the parent role.
    self.assertRaises(securesystemslib.exceptions.ForbiddenTargetError,
        securesystemslib.util.ensure_all_targets_allowed, 'targets/warehouse',
        ['file1.zip'], parent_delegations)

    self.assertRaises(securesystemslib.exceptions.ForbiddenTargetError,
        securesystemslib.util.ensure_all_targets_allowed, 'targets/warehouse',
        ['file1.txt', 'bad-README.txt'], parent_delegations)

    # Test for required attributes.
    # Missing 'paths' attribute.
    del parent_delegations['roles'][0]['paths']
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.ensure_all_targets_allowed,
        'targets/warehouse', list_of_targets, parent_delegations)

    # Test 'path_hash_prefixes' attribute.
    path_hash_prefixes = ['e3a3', '8fae', 'd543']
    parent_delegations['roles'][0]['path_hash_prefixes'] = path_hash_prefixes

    # Test normal case for 'path_hash_prefixes'.
    securesystemslib.util.ensure_all_targets_allowed('targets/warehouse',
        list_of_targets, parent_delegations)

    # Test target file with a path_hash_prefix that is not allowed in its
    # parent role.
    path_hash_prefix = securesystemslib.util.get_target_hash('file5.txt')[0:4]
    self.assertTrue(path_hash_prefix not in parent_delegations['roles'][0]
        ['path_hash_prefixes'])
    self.assertRaises(securesystemslib.exceptions.ForbiddenTargetError,
        securesystemslib.util.ensure_all_targets_allowed,
        'targets/warehouse', ['file5.txt'], parent_delegations)



  def test_C5_unittest_toolbox_make_temp_directory(self):
    # Verify that the tearDown function does not fail when
    # unittest_toolbox.make_temp_directory deletes the generated temp directory
    # here.
    temp_directory = self.make_temp_directory()
    os.rmdir(temp_directory)



  def test_c5_unittest_toolbox_random_path(self):
    # Verify that a random path can be generated with unittest_toolbox.
    random_path = self.random_path(length=10)
    self.assertTrue(securesystemslib.formats.PATH_SCHEMA.matches(random_path))
    self.assertTrue(10, len(random_path))


  def test_c6_get_compressed_length(self):
   self.temp_fileobj.write(b'hello world')
   self.assertTrue(self.temp_fileobj.get_compressed_length() == 11)

   temp_file = securesystemslib.util.TempFile()



  def test_digests_are_equal(self):
    digest = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'

    # Normal case: test for digests that are equal.
    self.assertTrue(securesystemslib.util.digests_are_equal(digest, digest))

    # Normal case: test for digests that are unequal.
    self.assertFalse(securesystemslib.util.digests_are_equal(digest, '0a8df1'))

    # Test for invalid arguments.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.digests_are_equal, 7, digest)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.util.digests_are_equal, digest, 7)

    # Test that digests_are_equal() takes the same amount of time to compare
    # equal and unequal arguments.
    runtime = timeit.timeit('digests_are_equal("ab8df", "ab8df")',
        setup='from securesystemslib.util import digests_are_equal',
        number=100000)

    runtime2 = timeit.timeit('digests_are_equal("ab8df", "1b8df")',
        setup='from securesystemslib.util import digests_are_equal',
        number=100000)

    runtime3 = timeit.timeit('"ab8df" == "ab8df"', number=100000)

    runtime4 = timeit.timeit('"ab8df" == "1b8df"', number=1000000)

    # The ratio for the 'digest_are_equal' runtimes should be at or near 1.
    ratio_digests_are_equal = abs(runtime2 / runtime)

    # The ratio for the variable-time runtimes should be (>1) & at or near 10?
    ratio_variable_compare = abs(runtime4 / runtime3)

    self.assertTrue(ratio_digests_are_equal < ratio_variable_compare)



# Run unit test.
if __name__ == '__main__':
  unittest.main()



# ed25519.py - Optimized version of the reference implementation of Ed25519
#
# Written in 2011? by Daniel J. Bernstein <djb@cr.yp.to>
#            2013 by Donald Stufft <donald@stufft.io>
#            2013 by Alex Gaynor <alex.gaynor@gmail.com>
#            2013 by Greg Price <price@mit.edu>
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain
# worldwide. This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along
# with this software. If not, see
# <http://creativecommons.org/publicdomain/zero/1.0/>.
import os
import timeit

import ed25519


seed = os.urandom(32)

data = b"The quick brown fox jumps over the lazy dog"
private_key = seed
public_key = ed25519.publickey_unsafe(seed)
signature = ed25519.signature_unsafe(data, private_key, public_key)

print('\nTime verify signature')
print(
    timeit.timeit(
        "ed25519.checkvalid(signature, data, public_key)",
        setup="from __main__ import ed25519, signature, data, public_key",
        number=100,
    )
)



#!/usr/bin/env python

"""
<Program Name>
  test_schema.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  October 2012.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Unit test for 'schema.py'
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import unittest
import re
import logging

import securesystemslib.exceptions
import securesystemslib.schema as SCHEMA

logger = logging.getLogger('securesystemslib.test_schema')


class TestSchema(unittest.TestCase):
  def setUp(self):
    pass



  def tearDown(self):
    pass



  def test_Schema(self):
    # Test conditions for the instantation of classes that inherit
    # from class Schema().
    class NewSchema(SCHEMA.Schema):
      def __init__(self):
        pass

    new_schema = NewSchema()
    self.assertRaises(NotImplementedError, new_schema.matches, 'test')

    # Define a new schema.
    class NewSchema2(SCHEMA.Schema):
      def __init__(self, string):
        self._string = string

      def check_match(self, object):
        if self._string != object:
          message = 'Expected: '+repr(self._string)
          raise securesystemslib.exceptions.FormatError(message)

    new_schema2 = NewSchema2('test')
    self.assertRaises(securesystemslib.exceptions.FormatError,
        new_schema2.check_match, 'bad')
    self.assertFalse(new_schema2.matches('bad'))
    self.assertTrue(new_schema2.matches('test'))

    # Test conditions for invalid arguments.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        new_schema2.check_match, True)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        new_schema2.check_match, NewSchema2)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        new_schema2.check_match, 123)

    self.assertFalse(new_schema2.matches(True))
    self.assertFalse(new_schema2.matches(NewSchema2))
    self.assertFalse(new_schema2.matches(123))



  def test_Any(self):
    # Test conditions for valid arguments.
    any_schema = SCHEMA.Any()

    self.assertTrue(any_schema.matches('test'))
    self.assertTrue(any_schema.matches(123))
    self.assertTrue(any_schema.matches(['test']))
    self.assertTrue(any_schema.matches({'word':'definition'}))
    self.assertTrue(any_schema.matches(True))



  def test_String(self):
    # Test conditions for valid arguments.
    string_schema = SCHEMA.String('test')

    self.assertTrue(string_schema.matches('test'))

    # Test conditions for invalid arguments.
    self.assertFalse(string_schema.matches(True))
    self.assertFalse(string_schema.matches(['test']))
    self.assertFalse(string_schema.matches(SCHEMA.Schema))

    # Test conditions for invalid arguments in a schema definition.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.String, 1)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.String, [1])
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.String, {'a': 1})



  def test_AnyString(self):
    # Test conditions for valid arguments.
    anystring_schema = SCHEMA.AnyString()

    self.assertTrue(anystring_schema.matches(''))
    self.assertTrue(anystring_schema.matches('a string'))

    # Test conditions for invalid arguments.
    self.assertFalse(anystring_schema.matches(['a']))
    self.assertFalse(anystring_schema.matches(3))
    self.assertFalse(anystring_schema.matches({'a': 'string'}))



  def test_OneOf(self):
    # Test conditions for valid arguments.
    oneof_schema = SCHEMA.OneOf([SCHEMA.ListOf(SCHEMA.Integer()),
        SCHEMA.String('Hello'), SCHEMA.String('bye')])

    self.assertTrue(oneof_schema.matches([]))
    self.assertTrue(oneof_schema.matches('bye'))
    self.assertTrue(oneof_schema.matches([1,2]))

    # Test conditions for invalid arguments.
    self.assertFalse(oneof_schema.matches(3))
    self.assertFalse(oneof_schema.matches(['Hi']))

    # Test conditions for invalid arguments in a schema definition.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.OneOf, 1)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.OneOf, [1])
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.OneOf, {'a': 1})
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.OneOf, [SCHEMA.AnyString(), 1])



  def test_AllOf(self):
    # Test conditions for valid arguments.
    allof_schema = SCHEMA.AllOf([SCHEMA.Any(),
        SCHEMA.AnyString(), SCHEMA.String('a')])

    self.assertTrue(allof_schema.matches('a'))

    # Test conditions for invalid arguments.
    self.assertFalse(allof_schema.matches('b'))

    # Test conditions for invalid arguments in a schema definition.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.AllOf, 1)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.AllOf, [1])
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.AllOf, {'a': 1})
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.AllOf, [SCHEMA.AnyString(), 1])


  def test_Boolean(self):
    # Test conditions for valid arguments.
    boolean_schema = SCHEMA.Boolean()

    self.assertTrue(boolean_schema.matches(True) and boolean_schema.matches(False))

    # Test conditions for invalid arguments.
    self.assertFalse(boolean_schema.matches(11))



  def test_ListOf(self):
    # Test conditions for valid arguments.
    listof_schema = SCHEMA.ListOf(SCHEMA.RegularExpression('(?:..)*'))
    listof2_schema = SCHEMA.ListOf(SCHEMA.Integer(),
        min_count=3, max_count=10)

    self.assertTrue(listof_schema.matches([]))
    self.assertTrue(listof_schema.matches(['Hi', 'this', 'list', 'is',
        'full', 'of', 'even', 'strs']))

    self.assertTrue(listof2_schema.matches([3]*3))
    self.assertTrue(listof2_schema.matches([3]*10))

    # Test conditions for invalid arguments.
    self.assertFalse(listof_schema.matches('hi'))
    self.assertFalse(listof_schema.matches({}))
    self.assertFalse(listof_schema.matches(['This', 'one', 'is not']))

    self.assertFalse(listof2_schema.matches([3]*2))
    self.assertFalse(listof2_schema.matches(([3]*11)))

    # Test conditions for invalid arguments in a schema definition.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.ListOf, 1)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.ListOf, [1])
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.ListOf, {'a': 1})



  def test_Integer(self):
    # Test conditions for valid arguments.
    integer_schema = SCHEMA.Integer()

    self.assertTrue(integer_schema.matches(99))
    self.assertTrue(SCHEMA.Integer(lo=10, hi=30).matches(25))

    # Test conditions for invalid arguments.
    self.assertFalse(integer_schema.matches(False))
    self.assertFalse(integer_schema.matches('a string'))
    self.assertFalse(SCHEMA.Integer(lo=10, hi=30).matches(5))



  def test_DictOf(self):
    # Test conditions for valid arguments.
    dictof_schema = SCHEMA.DictOf(SCHEMA.RegularExpression(r'[aeiou]+'),
        SCHEMA.Struct([SCHEMA.AnyString(), SCHEMA.AnyString()]))

    self.assertTrue(dictof_schema.matches({}))
    self.assertTrue(dictof_schema.matches({'a': ['x', 'y'], 'e' : ['', '']}))

    # Test conditions for invalid arguments.
    self.assertFalse(dictof_schema.matches(''))
    self.assertFalse(dictof_schema.matches({'a': ['x', 3], 'e' : ['', '']}))
    self.assertFalse(dictof_schema.matches({'a': ['x', 'y'], 'e' : ['', ''],
        'd' : ['a', 'b']}))

    # Test conditions for invalid arguments in a schema definition.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.DictOf, 1, 1)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.DictOf, [1], [1])
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.DictOf, {'a': 1}, 1)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.DictOf, SCHEMA.AnyString(), 1)



  def test_Optional(self):
    # Test conditions for valid arguments.
    optional_schema = SCHEMA.Object(k1=SCHEMA.String('X'),
        k2=SCHEMA.Optional(SCHEMA.String('Y')))

    self.assertTrue(optional_schema.matches({'k1': 'X', 'k2': 'Y'}))
    self.assertTrue(optional_schema.matches({'k1': 'X'}))

    # Test conditions for invalid arguments.
    self.assertFalse(optional_schema.matches({'k1': 'X', 'k2': 'Z'}))

    # Test conditions for invalid arguments in a schema definition.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Optional, 1)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Optional, [1])
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Optional, {'a': 1})



  def test_Object(self):
    # Test conditions for valid arguments.
    object_schema = SCHEMA.Object(a=SCHEMA.AnyString(),
        bc=SCHEMA.Struct([SCHEMA.Integer(), SCHEMA.Integer()]))

    self.assertTrue(object_schema.matches({'a':'ZYYY', 'bc':[5,9]}))
    self.assertTrue(object_schema.matches({'a':'ZYYY', 'bc':[5,9], 'xx':5}))

    # Test conditions for invalid arguments.
    self.assertFalse(object_schema.matches({'a':'ZYYY', 'bc':[5,9,3]}))
    self.assertFalse(object_schema.matches({'a':'ZYYY'}))

    # Test conditions for invalid arguments in a schema definition.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Object, a='a')
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Object, a=[1])
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Object, a=SCHEMA.AnyString(), b=1)

    # Test condition for invalid non-dict arguments.
    self.assertFalse(object_schema.matches([{'a':'XYZ'}]))
    self.assertFalse(object_schema.matches(8))



  def test_Struct(self):
    # Test conditions for valid arguments.
    struct_schema = SCHEMA.Struct([SCHEMA.ListOf(SCHEMA.AnyString()),
        SCHEMA.AnyString(), SCHEMA.String('X')])
    struct2_schema = SCHEMA.Struct([SCHEMA.String('X')], allow_more=True)
    struct3_schema = SCHEMA.Struct([SCHEMA.String('X'),
        SCHEMA.Integer()], [SCHEMA.Integer()])

    self.assertTrue(struct_schema.matches([[], 'Q', 'X']))

    self.assertTrue(struct2_schema.matches(['X']))
    self.assertTrue(struct2_schema.matches(['X', 'Y']))
    self.assertTrue(struct2_schema.matches(['X', ['Y', 'Z']]))

    self.assertTrue(struct3_schema.matches(['X', 3]))
    self.assertTrue(struct3_schema.matches(['X', 3, 9]))

    # Test conditions for invalid arguments.
    self.assertFalse(struct_schema.matches(False))
    self.assertFalse(struct_schema.matches('Foo'))
    self.assertFalse(struct_schema.matches([[], 'Q', 'D']))
    self.assertFalse(struct_schema.matches([[3], 'Q', 'X']))
    self.assertFalse(struct_schema.matches([[], 'Q', 'X', 'Y']))

    self.assertFalse(struct2_schema.matches([]))
    self.assertFalse(struct2_schema.matches([['X']]))

    self.assertFalse(struct3_schema.matches([]))
    self.assertFalse(struct3_schema.matches({}))
    self.assertFalse(struct3_schema.matches(['X']))
    self.assertFalse(struct3_schema.matches(['X', 3, 9, 11]))
    self.assertFalse(struct3_schema.matches(['X', 3, 'A']))

    # Test conditions for invalid arguments in a schema definition.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Struct, 1)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Struct, [1])
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Struct, {'a': 1})
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.Struct, [SCHEMA.AnyString(), 1])



  def test_RegularExpression(self):
    # Test conditions for valid arguments.
    # RegularExpression(pattern, modifiers, re_object, re_name).
    re_schema = SCHEMA.RegularExpression('h.*d')

    self.assertTrue(re_schema.matches('hello world'))

    # Provide a pattern that contains the trailing '$'
    re_schema_2 = SCHEMA.RegularExpression(pattern='abc$',
        modifiers=0, re_object=None, re_name='my_re')

    self.assertTrue(re_schema_2.matches('abc'))

    # Test for valid optional arguments.
    compiled_re = re.compile('^[a-z].*')
    re_schema_optional = SCHEMA.RegularExpression(pattern='abc',
        modifiers=0, re_object=compiled_re, re_name='my_re')
    self.assertTrue(re_schema_optional.matches('abc'))

    # Valid arguments, but the 'pattern' argument is unset (required if the
    # 're_object' is 'None'.)
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.RegularExpression, None, 0, None, None)

    # Valid arguments, 're_name' is unset, and 'pattern' is None.  An exception
    # is not raised, but 're_name' is set to 'pattern'.
    re_schema_optional = SCHEMA.RegularExpression(pattern=None,
        modifiers=0, re_object=compiled_re, re_name=None)

    self.assertTrue(re_schema_optional.matches('abc'))
    self.assertTrue(re_schema_optional._re_name == 'pattern')

    # Test conditions for invalid arguments.
    self.assertFalse(re_schema.matches('Hello World'))
    self.assertFalse(re_schema.matches('hello world!'))
    self.assertFalse(re_schema.matches([33, 'Hello']))

    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.RegularExpression, 8)



  def test_LengthString(self):
    # Test conditions for valid arguments.
    length_string = SCHEMA.LengthString(11)

    self.assertTrue(length_string.matches('Hello World'))
    self.assertTrue(length_string.matches('Hello Marty'))

    # Test conditions for invalid arguments.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.LengthString, 'hello')

    self.assertFalse(length_string.matches('hello'))
    self.assertFalse(length_string.matches(8))



  def test_LengthBytes(self):
    # Test conditions for valid arguments.
    length_bytes = SCHEMA.LengthBytes(11)

    self.assertTrue(length_bytes.matches(b'Hello World'))
    self.assertTrue(length_bytes.matches(b'Hello Marty'))

    # Test conditions for invalid arguments.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.LengthBytes, 'hello')
    self.assertRaises(securesystemslib.exceptions.FormatError,
        SCHEMA.LengthBytes, True)

    self.assertFalse(length_bytes.matches(b'hello'))
    self.assertFalse(length_bytes.matches(8))



  def test_AnyBytes(self):
    # Test conditions for valid arguments.
    anybytes_schema = SCHEMA.AnyBytes()

    self.assertTrue(anybytes_schema.matches(b''))
    self.assertTrue(anybytes_schema.matches(b'a string'))

    # Test conditions for invalid arguments.
    self.assertFalse(anybytes_schema.matches('a string'))
    self.assertFalse(anybytes_schema.matches(['a']))
    self.assertFalse(anybytes_schema.matches(3))
    self.assertFalse(anybytes_schema.matches({'a': 'string'}))


# Run the unit tests.
if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python

"""
<Program Name>
  test_hash.py

<Authors>
  Geremy Condra
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  Refactored March 1, 2012 (VLAD).  Based on a previous version of this module.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Unit test for 'hash.py'.
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import logging
import tempfile
import unittest

import securesystemslib.exceptions
import securesystemslib.hash

import six

logger = logging.getLogger('securesystemslib.test_hash')


if not 'hashlib' in securesystemslib.hash.SUPPORTED_LIBRARIES:
  logger.warn('Not testing hashlib: could not be imported.')


class TestHash(unittest.TestCase):

  def _run_with_all_hash_libraries(self, test_func):
    for lib in securesystemslib.hash.SUPPORTED_LIBRARIES:
      test_func(lib)


  def test_md5_update(self):
    self._run_with_all_hash_libraries(self._do_md5_update)


  def _do_md5_update(self, library):
    digest_object = securesystemslib.hash.digest('md5', library)
    self.assertEqual(digest_object.hexdigest(),
        'd41d8cd98f00b204e9800998ecf8427e')
    digest_object.update('a'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        '0cc175b9c0f1b6a831c399e269772661')
    digest_object.update('bbb'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        'f034e93091235adbb5d2781908e2b313')
    digest_object.update(''.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        'f034e93091235adbb5d2781908e2b313')


  def test_sha1_update(self):
    self._run_with_all_hash_libraries(self._do_sha1_update)


  def _do_sha1_update(self, library):
    digest_object = securesystemslib.hash.digest('sha1', library)

    self.assertEqual(digest_object.hexdigest(),
        'da39a3ee5e6b4b0d3255bfef95601890afd80709')
    digest_object.update('a'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        '86f7e437faa5a7fce15d1ddcb9eaeaea377667b8')
    digest_object.update('bbb'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        'd7bfa42fc62b697bf6cf1cda9af1fb7f40a27817')
    digest_object.update(''.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        'd7bfa42fc62b697bf6cf1cda9af1fb7f40a27817')


  def test_sha224_update(self):
    self._run_with_all_hash_libraries(self._do_sha224_update)


  def _do_sha224_update(self, library):
    digest_object = securesystemslib.hash.digest('sha224', library)

    self.assertEqual(digest_object.hexdigest(),
        'd14a028c2a3a2bc9476102bb288234c415a2b01f828ea62ac5b3e42f')
    digest_object.update('a'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        'abd37534c7d9a2efb9465de931cd7055ffdb8879563ae98078d6d6d5')
    digest_object.update('bbb'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        'ab1342f31c2a6f242d9a3cefb503fb49465c95eb255c16ad791d688c')
    digest_object.update(''.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        'ab1342f31c2a6f242d9a3cefb503fb49465c95eb255c16ad791d688c')


  def test_sha256_update(self):
    self._run_with_all_hash_libraries(self._do_sha256_update)


  def _do_sha256_update(self, library):
    digest_object = securesystemslib.hash.digest('sha256', library)
    self.assertEqual(digest_object.hexdigest(),
        'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855')
    digest_object.update('a'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb')
    digest_object.update('bbb'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        '01d162a5c95d4698c0a3e766ae80d85994b549b877ed275803725f43dadc83bd')
    digest_object.update(''.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
        '01d162a5c95d4698c0a3e766ae80d85994b549b877ed275803725f43dadc83bd')


  def test_sha384_update(self):
    self._run_with_all_hash_libraries(self._do_sha384_update)


  def _do_sha384_update(self, library):
    digest_object = securesystemslib.hash.digest('sha384', library)
    self.assertEqual(digest_object.hexdigest(),
    '38b060a751ac96384cd9327eb1b1e36a21fdb71114be07434c0cc7bf63f6e1da274edebfe'
    '76f65fbd51ad2f14898b95b')
    digest_object.update('a'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
    '54a59b9f22b0b80880d8427e548b7c23abd873486e1f035dce9cd697e85175033caa88e6d'
    '57bc35efae0b5afd3145f31')
    digest_object.update('bbb'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
    'f2c1438e9cc1d24bebbf3b88e60adc169db0c5c459d02054ec131438bf20ebee5ca88c17c'
    'b5f1a824fcccf8d2b20b0a9')
    digest_object.update(''.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
    'f2c1438e9cc1d24bebbf3b88e60adc169db0c5c459d02054ec131438bf20ebee5ca88c17c'
    'b5f1a824fcccf8d2b20b0a9')


  def test_sha512_update(self):
    self._run_with_all_hash_libraries(self._do_sha512_update)


  def _do_sha512_update(self, library):
    digest_object = securesystemslib.hash.digest('sha512', library)

    self.assertEqual(digest_object.hexdigest(),
    'cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5'
    'd85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e')
    digest_object.update('a'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
    '1f40fc92da241694750979ee6cf582f2d5d7d28e18335de05abc54d0560e0f5302860c652'
    'bf08d560252aa5e74210546f369fbbbce8c12cfc7957b2652fe9a75')
    digest_object.update('bbb'.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
    '09ade82ae3c5d54f8375f348563a372106488adef16a74b63b5591849f740bff55ceab22e'
    '117b4b09349b860f8a644adb32a9ea542abdecb80bf625160604251')
    digest_object.update(''.encode('utf-8'))
    self.assertEqual(digest_object.hexdigest(),
    '09ade82ae3c5d54f8375f348563a372106488adef16a74b63b5591849f740bff55ceab22e'
    '117b4b09349b860f8a644adb32a9ea542abdecb80bf625160604251')


  def test_unsupported_algorithm(self):
    self._run_with_all_hash_libraries(self._do_unsupported_algorithm)


  def _do_unsupported_algorithm(self, library):
    self.assertRaises(securesystemslib.exceptions.UnsupportedAlgorithmError,
        securesystemslib.hash.digest, 'bogus', library)


  def test_digest_size(self):
    self._run_with_all_hash_libraries(self._do_digest_size)


  def _do_digest_size(self, library):
    self.assertEqual(16,
        securesystemslib.hash.digest('md5', library).digest_size)
    self.assertEqual(20,
        securesystemslib.hash.digest('sha1', library).digest_size)
    self.assertEqual(28,
        securesystemslib.hash.digest('sha224', library).digest_size)
    self.assertEqual(32,
        securesystemslib.hash.digest('sha256', library).digest_size)
    self.assertEqual(48,
        securesystemslib.hash.digest('sha384', library).digest_size)
    self.assertEqual(64,
        securesystemslib.hash.digest('sha512', library).digest_size)


  def test_update_filename(self):
    self._run_with_all_hash_libraries(self._do_update_filename)


  def _do_update_filename(self, library):
    data = 'abcdefgh' * 4096
    fd, filename = tempfile.mkstemp()
    try:
      os.write(fd, data.encode('utf-8'))
      os.close(fd)
      for algorithm in ['md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512']:
        digest_object_truth = securesystemslib.hash.digest(algorithm, library)
        digest_object_truth.update(data.encode('utf-8'))
        digest_object = securesystemslib.hash.digest_filename(filename,
            algorithm, library)
        self.assertEqual(digest_object_truth.digest(), digest_object.digest())

    finally:
        os.remove(filename)


  def test_update_filename_normalize(self):
    self._run_with_all_hash_libraries(self._do_update_filename_normalize)


  def _do_update_filename_normalize(self, library):
    data = b'ab\r\nd\nf\r' * 4096
    normalized_data = data.replace(b'\r\n', b'\n').replace(b'\r', b'\n')
    fd, filename = tempfile.mkstemp()
    try:
      os.write(fd, data)
      os.close(fd)
      for algorithm in ['md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512']:
        digest_object_truth = securesystemslib.hash.digest(algorithm, library)
        digest_object_truth.update(normalized_data)
        digest_object = securesystemslib.hash.digest_filename(filename,
            algorithm, library, normalize_line_endings=True)
        self.assertEqual(digest_object_truth.digest(), digest_object.digest())

    finally:
      os.remove(filename)


  def test_update_file_obj(self):
    self._run_with_all_hash_libraries(self._do_update_file_obj)


  def _do_update_file_obj(self, library):
    data = 'abcdefgh' * 4096
    file_obj = six.StringIO()
    file_obj.write(data)
    for algorithm in ['md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512']:
      digest_object_truth = securesystemslib.hash.digest(algorithm, library)
      digest_object_truth.update(data.encode('utf-8'))
      digest_object = securesystemslib.hash.digest_fileobject(file_obj,
          algorithm, library)

      # Note: we don't seek because the update_file_obj call is supposed
      # to always seek to the beginning.
      self.assertEqual(digest_object_truth.digest(), digest_object.digest())


  def test_digest_from_rsa_scheme(self):
    self._run_with_all_hash_libraries(self._do_get_digest_from_rsa_valid_schemes)
    self._run_with_all_hash_libraries(self._do_get_digest_from_rsa_non_valid_schemes)


  def _do_get_digest_from_rsa_valid_schemes(self, library):
    algorithm = 'sha256'
    scheme = 'rsassa-pss-sha256'
    expected_digest_cls = type(securesystemslib.hash.digest(algorithm, library))

    self.assertIsInstance(securesystemslib.hash.digest_from_rsa_scheme(scheme, library),
      expected_digest_cls)

  def _do_get_digest_from_rsa_non_valid_schemes(self, library):
    self.assertRaises(securesystemslib.exceptions.FormatError,
      securesystemslib.hash.digest_from_rsa_scheme, 'rsassa-pss-sha123', library)



  def test_unsupported_digest_algorithm_and_library(self):
    self.assertRaises(securesystemslib.exceptions.UnsupportedAlgorithmError,
        securesystemslib.hash.digest, 'sha123', 'hashlib')
    self.assertRaises(securesystemslib.exceptions.UnsupportedLibraryError,
        securesystemslib.hash.digest, 'sha256', 'badlib')


# Run unit test.
if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python

"""
<Program Name>
  formats.py

<Author>
  Geremy Condra
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  Refactored April 30, 2012. -vladimir.v.diaz

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  A central location for all format-related checking of TUF objects.
  Note: 'formats.py' depends heavily on 'schema.py', so the 'schema.py'
  module should be read and understood before tackling this module.

  'formats.py' can be broken down into three sections.  (1) Schemas and object
  matching.  (2) Classes that represent Role Metadata and help produce
  correctly formatted files.  (3) Functions that help produce or verify TUF
  objects.

  The first section deals with schemas and object matching based on format.
  There are two ways of checking the format of objects.  The first method
  raises a 'securesystemslib.exceptions.FormatError' exception if the match
  fails and the other returns a Boolean result.

  securesystemslib.formats.<SCHEMA>.check_match(object)
  securesystemslib.formats.<SCHEMA>.matches(object)

  Example:

  rsa_key = {'keytype': 'rsa'
             'keyid': 34892fc465ac76bc3232fab
             'keyval': {'public': 'public_key',
                        'private': 'private_key'}

  securesystemslib.formats.RSAKEY_SCHEMA.check_match(rsa_key)
  securesystemslib.formats.RSAKEY_SCHEMA.matches(rsa_key)

  In this example, if a dict key or dict value is missing or incorrect,
  the match fails.  There are numerous variations of object checking
  provided by 'formats.py' and 'schema.py'.

  The second section deals with the role metadata classes.  There are
  multiple top-level roles, each with differing metadata formats.
  Example:

  root_object = securesystemslib.formats.RootFile.from_metadata(root_metadata_file)
  targets_metadata = securesystemslib.formats.TargetsFile.make_metadata(...)

  The input and output of these classes are checked against their respective
  schema to ensure correctly formatted metadata.

  The last section contains miscellaneous functions related to the format of
  TUF objects.
  Example:

  signable_object = make_signable(unsigned_object)
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import binascii
import calendar
import re
import string
import datetime
import time
import six

import securesystemslib.schema as SCHEMA
import securesystemslib.exceptions

# Note that in the schema definitions below, the 'SCHEMA.Object' types allow
# additional keys which are not defined. Thus, any additions to them will be
# easily backwards compatible with clients that are already deployed.

# A datetime in 'YYYY-MM-DDTHH:MM:SSZ' ISO 8601 format.  The "Z" zone designator
# for the zero UTC offset is always used (i.e., a numerical offset is not
# supported.)  Example: '2015-10-21T13:20:00Z'.  Note:  This is a simple format
# check, and an ISO8601 string should be fully verified when it is parsed.
ISO8601_DATETIME_SCHEMA = SCHEMA.RegularExpression(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z')

# A Unix/POSIX time format.  An integer representing the number of seconds
# since the epoch (January 1, 1970.)  Metadata uses this format for the
# 'expires' field.  Set 'hi' to the upper timestamp limit (year 2038), the max
# value of an int.
UNIX_TIMESTAMP_SCHEMA = SCHEMA.Integer(lo=0, hi=2147483647)

# A hexadecimal value in '23432df87ab..' format.
HASH_SCHEMA = SCHEMA.RegularExpression(r'[a-fA-F0-9]+')

# A dict in {'sha256': '23432df87ab..', 'sha512': '34324abc34df..', ...} format.
HASHDICT_SCHEMA = SCHEMA.DictOf(
  key_schema = SCHEMA.AnyString(),
  value_schema = HASH_SCHEMA)

# A hexadecimal value in '23432df87ab..' format.
HEX_SCHEMA = SCHEMA.RegularExpression(r'[a-fA-F0-9]+')

# A key identifier (e.g., a hexadecimal value identifying an RSA key).
KEYID_SCHEMA = HASH_SCHEMA

# A list of KEYID_SCHEMA.
KEYIDS_SCHEMA = SCHEMA.ListOf(KEYID_SCHEMA)

# The signing scheme used by a key to generate a signature (e.g.,
# 'rsassa-pss-sha256' is one of the signing schemes for key type 'rsa').
SCHEME_SCHEMA = SCHEMA.AnyString()

# A relative file path (e.g., 'metadata/root/').
RELPATH_SCHEMA = SCHEMA.AnyString()
RELPATHS_SCHEMA = SCHEMA.ListOf(RELPATH_SCHEMA)

# An absolute path.
PATH_SCHEMA = SCHEMA.AnyString()
PATHS_SCHEMA = SCHEMA.ListOf(PATH_SCHEMA)

# Uniform Resource Locator identifier (e.g., 'https://www.updateframework.com/').
URL_SCHEMA = SCHEMA.AnyString()

# A dictionary holding version information.
VERSION_SCHEMA = SCHEMA.Object(
  object_name = 'VERSION_SCHEMA',
  major = SCHEMA.Integer(lo=0),
  minor = SCHEMA.Integer(lo=0),
  fix = SCHEMA.Integer(lo=0))

# An integer representing the numbered version of a metadata file.
# Must be 1, or greater.
METADATAVERSION_SCHEMA = SCHEMA.Integer(lo=0)

# An integer representing length.  Must be 0, or greater.
LENGTH_SCHEMA = SCHEMA.Integer(lo=0)

# An integer representing logger levels, such as logging.CRITICAL (=50).
# Must be between 0 and 50.
LOGLEVEL_SCHEMA = SCHEMA.Integer(lo=0, hi=50)

# A string representing a named object.
NAME_SCHEMA = SCHEMA.AnyString()
NAMES_SCHEMA = SCHEMA.ListOf(NAME_SCHEMA)

# A byte string representing data.
DATA_SCHEMA = SCHEMA.AnyBytes()

# A text string.  For instance, a string entered by the user.
TEXT_SCHEMA = SCHEMA.AnyString()

# Supported hash algorithms.
HASHALGORITHMS_SCHEMA = SCHEMA.ListOf(SCHEMA.OneOf(
  [SCHEMA.String('md5'), SCHEMA.String('sha1'),
   SCHEMA.String('sha224'), SCHEMA.String('sha256'),
   SCHEMA.String('sha384'), SCHEMA.String('sha512')]))

# The contents of an encrypted TUF key.  Encrypted TUF keys are saved to files
# in this format.
ENCRYPTEDKEY_SCHEMA = SCHEMA.AnyString()

# A value that is either True or False, on or off, etc.
BOOLEAN_SCHEMA = SCHEMA.Boolean()

# A role's threshold value (i.e., the minimum number
# of signatures required to sign a metadata file).
# Must be 1 and greater.
THRESHOLD_SCHEMA = SCHEMA.Integer(lo=1)

# A string representing a role's name.
ROLENAME_SCHEMA = SCHEMA.AnyString()

# The minimum number of bits for an RSA key.  Must be 2048 bits, or greater
# (recommended by TUF).  Recommended RSA key sizes:
# http://www.emc.com/emc-plus/rsa-labs/historical/twirl-and-rsa-key-size.htm#table1
RSAKEYBITS_SCHEMA = SCHEMA.Integer(lo=2048)

# The supported ECDSA signature schemes (ecdsa-sha2-nistp256 is supported by
# default).
ECDSA_SCHEME_SCHEMA = SCHEMA.OneOf([SCHEMA.String('ecdsa-sha2-nistp256')])

# The number of hashed bins, or the number of delegated roles.  See
# delegate_hashed_bins() in 'repository_tool.py' for an example.  Note:
# Tools may require further restrictions on the number of bins, such
# as requiring them to be a power of 2.
NUMBINS_SCHEMA = SCHEMA.Integer(lo=1)

# A pyca-cryptography signature.
PYCACRYPTOSIGNATURE_SCHEMA = SCHEMA.AnyBytes()

# An RSA key in PEM format.
PEMRSA_SCHEMA = SCHEMA.AnyString()

# An ECDSA key in PEM format.
PEMECDSA_SCHEMA = SCHEMA.AnyString()

# A string representing a password.
PASSWORD_SCHEMA = SCHEMA.AnyString()

# A list of passwords.
PASSWORDS_SCHEMA = SCHEMA.ListOf(PASSWORD_SCHEMA)

# The actual values of a key, as opposed to meta data such as a key type and
# key identifier ('rsa', 233df889cb).  For RSA keys, the key value is a pair of
# public and private keys in PEM Format stored as strings.
KEYVAL_SCHEMA = SCHEMA.Object(
  object_name = 'KEYVAL_SCHEMA',
  public = SCHEMA.AnyString(),
  private = SCHEMA.Optional(SCHEMA.AnyString()))

# Public keys CAN have a private portion (for backwards compatibility) which
# MUST be an empty string
PUBLIC_KEYVAL_SCHEMA = SCHEMA.Object(
  object_name = 'KEYVAL_SCHEMA',
  public = SCHEMA.AnyString(),
  private = SCHEMA.Optional(SCHEMA.String("")))

# Supported TUF key types.
KEYTYPE_SCHEMA = SCHEMA.OneOf(
  [SCHEMA.String('rsa'), SCHEMA.String('ed25519'),
   SCHEMA.String('ecdsa-sha2-nistp256')])

# A generic TUF key.  All TUF keys should be saved to metadata files in this
# format.
KEY_SCHEMA = SCHEMA.Object(
  object_name = 'KEY_SCHEMA',
  keytype = SCHEMA.AnyString(),
  scheme = SCHEME_SCHEMA,
  keyval = KEYVAL_SCHEMA,
  expires = SCHEMA.Optional(ISO8601_DATETIME_SCHEMA))

# Like KEY_SCHEMA, but requires keyval's private portion to be unset or empty,
# and optionally includes the supported keyid hash algorithms used to generate
# the key's keyid.
PUBLIC_KEY_SCHEMA = SCHEMA.Object(
  object_name = 'PUBLIC_KEY_SCHEMA',
  keytype = SCHEMA.AnyString(),
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = PUBLIC_KEYVAL_SCHEMA,
  expires = SCHEMA.Optional(ISO8601_DATETIME_SCHEMA))

# A TUF key object.  This schema simplifies validation of keys that may be one
# of the supported key types.  Supported key types: 'rsa', 'ed25519'.
ANYKEY_SCHEMA = SCHEMA.Object(
  object_name = 'ANYKEY_SCHEMA',
  keytype = KEYTYPE_SCHEMA,
  scheme = SCHEME_SCHEMA,
  keyid = KEYID_SCHEMA,
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = KEYVAL_SCHEMA,
  expires = SCHEMA.Optional(ISO8601_DATETIME_SCHEMA))

# A list of TUF key objects.
ANYKEYLIST_SCHEMA = SCHEMA.ListOf(ANYKEY_SCHEMA)

# RSA signature schemes.
RSA_SCHEME_SCHEMA = SCHEMA.OneOf([
  SCHEMA.RegularExpression(r'rsassa-pss-(md5|sha1|sha224|sha256|sha384|sha512)'),
  SCHEMA.RegularExpression(r'rsa-pkcs1v15-(md5|sha1|sha224|sha256|sha384|sha512)')])

# An RSA TUF key.
RSAKEY_SCHEMA = SCHEMA.Object(
  object_name = 'RSAKEY_SCHEMA',
  keytype = SCHEMA.String('rsa'),
  scheme = RSA_SCHEME_SCHEMA,
  keyid = KEYID_SCHEMA,
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = KEYVAL_SCHEMA)

# An ECDSA TUF key.
ECDSAKEY_SCHEMA = SCHEMA.Object(
  object_name = 'ECDSAKEY_SCHEMA',
  keytype = SCHEMA.String('ecdsa-sha2-nistp256'),
  scheme = ECDSA_SCHEME_SCHEMA,
  keyid = KEYID_SCHEMA,
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = KEYVAL_SCHEMA)

# An ED25519 raw public key, which must be 32 bytes.
ED25519PUBLIC_SCHEMA = SCHEMA.LengthBytes(32)

# An ED25519 raw seed key, which must be 32 bytes.
ED25519SEED_SCHEMA = SCHEMA.LengthBytes(32)

# An ED25519 raw signature, which must be 64 bytes.
ED25519SIGNATURE_SCHEMA = SCHEMA.LengthBytes(64)

# An ECDSA signature.
ECDSASIGNATURE_SCHEMA = SCHEMA.AnyBytes()

# Required installation libraries expected by the repository tools and other
# cryptography modules.
REQUIRED_LIBRARIES_SCHEMA = SCHEMA.ListOf(SCHEMA.OneOf(
  [SCHEMA.String('general'), SCHEMA.String('ed25519'), SCHEMA.String('rsa'),
   SCHEMA.String('ecdsa-sha2-nistp256')]))

# Ed25519 signature schemes.  The vanilla Ed25519 signature scheme is currently
# supported.
ED25519_SIG_SCHEMA = SCHEMA.OneOf([SCHEMA.String('ed25519')])

# An ed25519 TUF key.
ED25519KEY_SCHEMA = SCHEMA.Object(
  object_name = 'ED25519KEY_SCHEMA',
  keytype = SCHEMA.String('ed25519'),
  scheme = ED25519_SIG_SCHEMA,
  keyid = KEYID_SCHEMA,
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = KEYVAL_SCHEMA)

# Information about target files, like file length and file hash(es).  This
# schema allows the storage of multiple hashes for the same file (e.g., sha256
# and sha512 may be computed for the same file and stored).
FILEINFO_SCHEMA = SCHEMA.Object(
  object_name = 'FILEINFO_SCHEMA',
  length = LENGTH_SCHEMA,
  hashes = HASHDICT_SCHEMA,
  version = SCHEMA.Optional(METADATAVERSION_SCHEMA),
  custom = SCHEMA.Optional(SCHEMA.Object()))

# Version information specified in "snapshot.json" for each role available on
# the TUF repository.  The 'FILEINFO_SCHEMA' object was previously listed in
# the snapshot role, but was switched to this object format to reduce the
# amount of metadata that needs to be downloaded.  Listing version numbers in
# "snapshot.json" also prevents rollback attacks for roles that clients have
# not downloaded.
VERSIONINFO_SCHEMA = SCHEMA.Object(
  object_name = 'VERSIONINFO_SCHEMA',
  version = METADATAVERSION_SCHEMA)

# A dict holding the version information for a particular metadata role.  The
# dict keys hold the relative file paths, and the dict values the corresponding
# version numbers.
VERSIONDICT_SCHEMA = SCHEMA.DictOf(
  key_schema = RELPATH_SCHEMA,
  value_schema = VERSIONINFO_SCHEMA)

# A dict holding the information for a particular target / file.  The dict keys
# hold the relative file paths, and the dict values the corresponding file
# information.
FILEDICT_SCHEMA = SCHEMA.DictOf(
  key_schema = RELPATH_SCHEMA,
  value_schema = FILEINFO_SCHEMA)

# A single signature of an object.  Indicates the signature, and the KEYID of
# the signing key.  I debated making the signature schema not contain the key
# ID and instead have the signatures of a file be a dictionary with the key
# being the keyid and the value being the signature schema without the keyid.
# That would be under the argument that a key should only be able to sign a
# file once.
SIGNATURE_SCHEMA = SCHEMA.Object(
  object_name = "SIGNATURE_SCHEMA",
  keyid = KEYID_SCHEMA,
  sig = HEX_SCHEMA)

# List of SIGNATURE_SCHEMA.
SIGNATURES_SCHEMA = SCHEMA.ListOf(SIGNATURE_SCHEMA)

# A schema holding the result of checking the signatures of a particular
# 'SIGNABLE_SCHEMA' role.
# For example, how many of the signatures for the 'Target' role are
# valid?  This SCHEMA holds this information.  See 'sig.py' for
# more information.
SIGNATURESTATUS_SCHEMA = SCHEMA.Object(
  object_name = 'SIGNATURESTATUS_SCHEMA',
  threshold = SCHEMA.Integer(),
  good_sigs = KEYIDS_SCHEMA,
  bad_sigs = KEYIDS_SCHEMA,
  unknown_sigs = KEYIDS_SCHEMA,
  untrusted_sigs = KEYIDS_SCHEMA)

# A signable object.  Holds the signing role and its associated signatures.
SIGNABLE_SCHEMA = SCHEMA.Object(
  object_name = 'SIGNABLE_SCHEMA',
  signed = SCHEMA.Any(),
  signatures = SCHEMA.ListOf(SIGNATURE_SCHEMA))

# A dict where the dict keys hold a keyid and the dict values a key object.
KEYDICT_SCHEMA = SCHEMA.DictOf(
  key_schema = KEYID_SCHEMA,
  value_schema = KEY_SCHEMA)

# The format used by the key database to store keys.  The dict keys hold a key
# identifier and the dict values any object.  The key database should store
# key objects in the values (e.g., 'RSAKEY_SCHEMA', 'DSAKEY_SCHEMA').
KEYDB_SCHEMA = SCHEMA.DictOf(
  key_schema = KEYID_SCHEMA,
  value_schema = SCHEMA.Any())

# A path hash prefix is a hexadecimal string.
PATH_HASH_PREFIX_SCHEMA = HEX_SCHEMA

# A list of path hash prefixes.
PATH_HASH_PREFIXES_SCHEMA = SCHEMA.ListOf(PATH_HASH_PREFIX_SCHEMA)

# Role object in {'keyids': [keydids..], 'name': 'ABC', 'threshold': 1,
# 'paths':[filepaths..]} format.
ROLE_SCHEMA = SCHEMA.Object(
  object_name = 'ROLE_SCHEMA',
  name = SCHEMA.Optional(ROLENAME_SCHEMA),
  keyids = KEYIDS_SCHEMA,
  threshold = THRESHOLD_SCHEMA,
  backtrack = SCHEMA.Optional(BOOLEAN_SCHEMA),
  paths = SCHEMA.Optional(RELPATHS_SCHEMA),
  path_hash_prefixes = SCHEMA.Optional(PATH_HASH_PREFIXES_SCHEMA))

# A dict of roles where the dict keys are role names and the dict values holding
# the role data/information.
ROLEDICT_SCHEMA = SCHEMA.DictOf(
  key_schema = ROLENAME_SCHEMA,
  value_schema = ROLE_SCHEMA)

# Like ROLEDICT_SCHEMA, except that ROLE_SCHEMA instances are stored in order.
ROLELIST_SCHEMA = SCHEMA.ListOf(ROLE_SCHEMA)

# The delegated roles of a Targets role (a parent).
DELEGATIONS_SCHEMA = SCHEMA.Object(
  keys = KEYDICT_SCHEMA,
  roles = ROLELIST_SCHEMA)

# The fileinfo format of targets specified in the repository and
# developer tools.  The second element of this list holds custom data about the
# target, such as file permissions, author(s), last modified, etc.
CUSTOM_SCHEMA = SCHEMA.Object()

PATH_FILEINFO_SCHEMA = SCHEMA.DictOf(
  key_schema = RELPATH_SCHEMA,
  value_schema = CUSTOM_SCHEMA)

# TUF roledb
ROLEDB_SCHEMA = SCHEMA.Object(
  object_name = 'ROLEDB_SCHEMA',
  keyids = SCHEMA.Optional(KEYIDS_SCHEMA),
  signing_keyids = SCHEMA.Optional(KEYIDS_SCHEMA),
  previous_keyids = SCHEMA.Optional(KEYIDS_SCHEMA),
  threshold = SCHEMA.Optional(THRESHOLD_SCHEMA),
  previous_threshold = SCHEMA.Optional(THRESHOLD_SCHEMA),
  version = SCHEMA.Optional(METADATAVERSION_SCHEMA),
  expires = SCHEMA.Optional(ISO8601_DATETIME_SCHEMA),
  signatures = SCHEMA.Optional(SIGNATURES_SCHEMA),
  paths = SCHEMA.Optional(SCHEMA.OneOf([RELPATHS_SCHEMA, PATH_FILEINFO_SCHEMA])),
  path_hash_prefixes = SCHEMA.Optional(PATH_HASH_PREFIXES_SCHEMA),
  delegations = SCHEMA.Optional(DELEGATIONS_SCHEMA),
  partial_loaded = SCHEMA.Optional(BOOLEAN_SCHEMA))

# Root role: indicates root keys and top-level roles.
ROOT_SCHEMA = SCHEMA.Object(
  object_name = 'ROOT_SCHEMA',
  _type = SCHEMA.String('root'),
  version = METADATAVERSION_SCHEMA,
  consistent_snapshot = BOOLEAN_SCHEMA,
  expires = ISO8601_DATETIME_SCHEMA,
  keys = KEYDICT_SCHEMA,
  roles = ROLEDICT_SCHEMA)

# Targets role: Indicates targets and delegates target paths to other roles.
TARGETS_SCHEMA = SCHEMA.Object(
  object_name = 'TARGETS_SCHEMA',
  _type = SCHEMA.String('targets'),
  version = METADATAVERSION_SCHEMA,
  expires = ISO8601_DATETIME_SCHEMA,
  targets = FILEDICT_SCHEMA,
  delegations = SCHEMA.Optional(DELEGATIONS_SCHEMA))

# Snapshot role: indicates the latest versions of all metadata (except
# timestamp).
SNAPSHOT_SCHEMA = SCHEMA.Object(
  object_name = 'SNAPSHOT_SCHEMA',
  _type = SCHEMA.String('snapshot'),
  version = METADATAVERSION_SCHEMA,
  expires = ISO8601_DATETIME_SCHEMA,
  meta = VERSIONDICT_SCHEMA)

# Timestamp role: indicates the latest version of the snapshot file.
TIMESTAMP_SCHEMA = SCHEMA.Object(
  object_name = 'TIMESTAMP_SCHEMA',
  _type = SCHEMA.String('timestamp'),
  version = METADATAVERSION_SCHEMA,
  expires = ISO8601_DATETIME_SCHEMA,
  meta = FILEDICT_SCHEMA)

# project.cfg file: stores information about the project in a json dictionary
PROJECT_CFG_SCHEMA = SCHEMA.Object(
    object_name = 'PROJECT_CFG_SCHEMA',
    project_name = SCHEMA.AnyString(),
    layout_type = SCHEMA.OneOf([SCHEMA.String('repo-like'), SCHEMA.String('flat')]),
    targets_location = PATH_SCHEMA,
    metadata_location = PATH_SCHEMA,
    prefix = PATH_SCHEMA,
    public_keys = KEYDICT_SCHEMA,
    threshold = SCHEMA.Integer(lo = 0, hi = 2)
    )

# A schema containing information a repository mirror may require,
# such as a url, the path of the directory metadata files, etc.
MIRROR_SCHEMA = SCHEMA.Object(
  object_name = 'MIRROR_SCHEMA',
  url_prefix = URL_SCHEMA,
  metadata_path = RELPATH_SCHEMA,
  targets_path = RELPATH_SCHEMA,
  confined_target_dirs = RELPATHS_SCHEMA,
  custom = SCHEMA.Optional(SCHEMA.Object()))

# A dictionary of mirrors where the dict keys hold the mirror's name and
# and the dict values the mirror's data (i.e., 'MIRROR_SCHEMA').
# The repository class of 'updater.py' accepts dictionaries
# of this type provided by the TUF client.
MIRRORDICT_SCHEMA = SCHEMA.DictOf(
  key_schema = SCHEMA.AnyString(),
  value_schema = MIRROR_SCHEMA)

# A Mirrorlist: indicates all the live mirrors, and what documents they
# serve.
MIRRORLIST_SCHEMA = SCHEMA.Object(
  object_name = 'MIRRORLIST_SCHEMA',
  _type = SCHEMA.String('mirrors'),
  version = METADATAVERSION_SCHEMA,
  expires = ISO8601_DATETIME_SCHEMA,
  mirrors = SCHEMA.ListOf(MIRROR_SCHEMA))

# Any of the role schemas (e.g., TIMESTAMP_SCHEMA, SNAPSHOT_SCHEMA, etc.)
ANYROLE_SCHEMA = SCHEMA.OneOf([ROOT_SCHEMA, TARGETS_SCHEMA, SNAPSHOT_SCHEMA,
                               TIMESTAMP_SCHEMA, MIRROR_SCHEMA])



def datetime_to_unix_timestamp(datetime_object):
  """
  <Purpose>
    Convert 'datetime_object' (in datetime.datetime()) format) to a Unix/POSIX
    timestamp.  For example, Python's time.time() returns a Unix timestamp, and
    includes the number of microseconds.  'datetime_object' is converted to UTC.

    >>> datetime_object = datetime.datetime(1985, 10, 26, 1, 22)
    >>> timestamp = datetime_to_unix_timestamp(datetime_object)
    >>> timestamp
    499137720

  <Arguments>
    datetime_object:
      The datetime.datetime() object to convert to a Unix timestamp.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'datetime_object' is not a
    datetime.datetime() object.

  <Side Effects>
    None.

  <Returns>
    A unix (posix) timestamp (e.g., 499137660).
  """

  # Is 'datetime_object' a datetime.datetime() object?
  # Raise 'securesystemslib.exceptions.FormatError' if not.
  if not isinstance(datetime_object, datetime.datetime):
    message = repr(datetime_object) + ' is not a datetime.datetime() object.'
    raise securesystemslib.exceptions.FormatError(message)

  unix_timestamp = calendar.timegm(datetime_object.timetuple())

  return unix_timestamp




def unix_timestamp_to_datetime(unix_timestamp):
  """
  <Purpose>
    Convert 'unix_timestamp' (i.e., POSIX time, in UNIX_TIMESTAMP_SCHEMA format)
    to a datetime.datetime() object.  'unix_timestamp' is the number of seconds
    since the epoch (January 1, 1970.)

    >>> datetime_object = unix_timestamp_to_datetime(1445455680)
    >>> datetime_object
    datetime.datetime(2015, 10, 21, 19, 28)

  <Arguments>
    unix_timestamp:
      An integer representing the time (e.g., 1445455680).  Conformant to
      'securesystemslib.formats.UNIX_TIMESTAMP_SCHEMA'.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'unix_timestamp' is improperly
    formatted.

  <Side Effects>
    None.

  <Returns>
    A datetime.datetime() object corresponding to 'unix_timestamp'.
  """

  # Is 'unix_timestamp' properly formatted?
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  securesystemslib.formats.UNIX_TIMESTAMP_SCHEMA.check_match(unix_timestamp)

  # Convert 'unix_timestamp' to a 'time.struct_time',  in UTC.  The Daylight
  # Savings Time (DST) flag is set to zero.  datetime.fromtimestamp() is not
  # used because it returns a local datetime.
  struct_time = time.gmtime(unix_timestamp)

  # Extract the (year, month, day, hour, minutes, seconds) arguments for the
  # datetime object to be returned.
  datetime_object = datetime.datetime(*struct_time[:6])

  return datetime_object




def format_base64(data):
  """
  <Purpose>
    Return the base64 encoding of 'data' with whitespace and '=' signs omitted.

  <Arguments>
    data:
      Binary or buffer of data to convert.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the base64 encoding fails or the
    argument is invalid.

  <Side Effects>
    None.

  <Returns>
    A base64-encoded string.
  """

  try:
    return binascii.b2a_base64(data).decode('utf-8').rstrip('=\n ')

  except (TypeError, binascii.Error) as e:
    raise securesystemslib.exceptions.FormatError('Invalid base64'
      ' encoding: ' + str(e))




def parse_base64(base64_string):
  """
  <Purpose>
    Parse a base64 encoding with whitespace and '=' signs omitted.

  <Arguments>
    base64_string:
      A string holding a base64 value.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'base64_string' cannot be parsed
    due to an invalid base64 encoding.

  <Side Effects>
    None.

  <Returns>
    A byte string representing the parsed based64 encoding of
    'base64_string'.
  """

  if not isinstance(base64_string, six.string_types):
    message = 'Invalid argument: '+repr(base64_string)
    raise securesystemslib.exceptions.FormatError(message)

  extra = len(base64_string) % 4
  if extra:
    padding = '=' * (4 - extra)
    base64_string = base64_string + padding

  try:
    return binascii.a2b_base64(base64_string.encode('utf-8'))

  except (TypeError, binascii.Error) as e:
    raise securesystemslib.exceptions.FormatError('Invalid base64'
      ' encoding: ' + str(e))




def _canonical_string_encoder(string):
  """
  <Purpose>
    Encode 'string' to canonical string format.

  <Arguments>
    string:
      The string to encode.

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    A string with the canonical-encoded 'string' embedded.
  """

  string = '"%s"' % re.sub(r'(["\\])', r'\\\1', string)

  return string


def _encode_canonical(object, output_function):
  # Helper for encode_canonical.  Older versions of json.encoder don't
  # even let us replace the separators.

  if isinstance(object, six.string_types):
    output_function(_canonical_string_encoder(object))
  elif object is True:
    output_function("true")
  elif object is False:
    output_function("false")
  elif object is None:
    output_function("null")
  elif isinstance(object, six.integer_types):
    output_function(str(object))
  elif isinstance(object, (tuple, list)):
    output_function("[")
    if len(object):
      for item in object[:-1]:
        _encode_canonical(item, output_function)
        output_function(",")
      _encode_canonical(object[-1], output_function)
    output_function("]")
  elif isinstance(object, dict):
    output_function("{")
    if len(object):
      items = sorted(six.iteritems(object))
      for key, value in items[:-1]:
        output_function(_canonical_string_encoder(key))
        output_function(":")
        _encode_canonical(value, output_function)
        output_function(",")
      key, value = items[-1]
      output_function(_canonical_string_encoder(key))
      output_function(":")
      _encode_canonical(value, output_function)
    output_function("}")
  else:
    raise securesystemslib.exceptions.FormatError('I cannot encode '+repr(object))


def encode_canonical(object, output_function=None):
  """
  <Purpose>
    Encode 'object' in canonical JSON form, as specified at
    http://wiki.laptop.org/go/Canonical_JSON .  It's a restricted
    dialect of JSON in which keys are always lexically sorted,
    there is no whitespace, floats aren't allowed, and only quote
    and backslash get escaped.  The result is encoded in UTF-8,
    and the resulting bits are passed to output_function (if provided),
    or joined into a string and returned.

    Note: This function should be called prior to computing the hash or
    signature of a JSON object in TUF.  For example, generating a signature
    of a signing role object such as 'ROOT_SCHEMA' is required to ensure
    repeatable hashes are generated across different json module versions
    and platforms.  Code elsewhere is free to dump JSON objects in any format
    they wish (e.g., utilizing indentation and single quotes around object
    keys).  These objects are only required to be in "canonical JSON" format
    when their hashes or signatures are needed.

    >>> encode_canonical("")
    '""'
    >>> encode_canonical([1, 2, 3])
    '[1,2,3]'
    >>> encode_canonical([])
    '[]'
    >>> encode_canonical({"A": [99]})
    '{"A":[99]}'
    >>> encode_canonical({"x" : 3, "y" : 2})
    '{"x":3,"y":2}'

  <Arguments>
    object:
      The object to be encoded.

    output_function:
      The result will be passed as arguments to 'output_function'
      (e.g., output_function('result')).

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'object' cannot be encoded or
    'output_function' is not callable.

  <Side Effects>
    The results are fed to 'output_function()' if 'output_function' is set.

  <Returns>
    A string representing the 'object' encoded in canonical JSON form.
  """

  result = None
  # If 'output_function' is unset, treat it as
  # appending to a list.
  if output_function is None:
    result = []
    output_function = result.append

  try:
    _encode_canonical(object, output_function)

  except (TypeError, securesystemslib.exceptions.FormatError) as e:
    message = 'Could not encode ' + repr(object) + ': ' + str(e)
    raise securesystemslib.exceptions.FormatError(message)

  # Return the encoded 'object' as a string.
  # Note: Implies 'output_function' is None,
  # otherwise results are sent to 'output_function'.
  if result is not None:
    return ''.join(result)
